# Copyright 2008-2019 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""This module contains tests for bodhi.client.bindings."""

from datetime import datetime, timedelta
from unittest import mock
import copy
import os

from requests import HTTPError
import munch
import pytest

from bodhi.client import bindings, constants
from bodhi.client.oidcclient import OIDCClientError

from . import fixtures as client_test_data
from .utils import build_response, compare_output


class BodhiClientTestCase:
    def setup_method(self, method):
        self.oidcclient_patcher = mock.patch("bodhi.client.bindings.OIDCClient")
        self.oidcclient_class = self.oidcclient_patcher.start()

    def teardown_method(self, method):
        self.oidcclient_patcher.stop()


class TestBodhiClientBase(BodhiClientTestCase):

    def test_base_url_not_ends_in_slash(self):
        """
        If the base_url doesn't end in a slash, __init__() should append one.
        """
        client = bindings.BodhiClient(base_url='http://localhost:6543')

        assert client.base_url == 'http://localhost:6543/'

    def test_id_provider(self, mocker):
        """Test the id_provider parameter."""
        storage = mocker.patch("bodhi.client.bindings.JSONStorage")
        storage.return_value = mocker.Mock()

        client = bindings.BodhiClient(
            base_url='http://example.com/bodhi/', client_id='CLIENT_ID',
            id_provider='https://id.example.com/', staging=False)

        assert client.base_url == 'http://example.com/bodhi/'
        self.oidcclient_class.assert_called_with(
            "CLIENT_ID",
            constants.SCOPE,
            "https://id.example.com",
            storage=storage.return_value,
        )
        storage.assert_called_with(os.path.expanduser("~/.config/bodhi/client.json"))
        assert client.csrf_token == ''

    def test_staging_true(self, mocker):
        """
        Test with staging set to True.
        """
        storage = mocker.patch("bodhi.client.bindings.JSONStorage")
        storage.return_value = mocker.Mock()

        client = bindings.BodhiClient(
            base_url='http://example.com/bodhi/', client_id='CLIENT_ID',
            id_provider='https://id.example.com/', staging=True)

        assert client.base_url == constants.STG_BASE_URL
        self.oidcclient_class.assert_called_with(
            constants.STG_CLIENT_ID,
            constants.SCOPE,
            constants.STG_IDP,
            storage=storage.return_value
        )
        assert client.csrf_token == ''


class TestBodhiClientAuth(BodhiClientTestCase):
    def setup_method(self, method):
        super().setup_method(method)
        self.oidc = mock.Mock(name="oidc")
        self.oidcclient_class.return_value = self.oidc

    def test_ensure_auth_ok(self):
        self.oidc.has_cookie.return_value = False
        self.oidc.request.return_value = build_response(200, "/login-token", "")
        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        client.ensure_auth()

        self.oidc.ensure_auth.assert_called_once_with(use_kerberos=True)
        self.oidc.request.assert_called_once_with(
            "GET", "http://example.com/bodhi/oidc/login-token"
        )
        self.oidc.login.assert_not_called()

    def test_ensure_auth_retry(self):
        self.oidc.has_cookie.return_value = False
        self.oidc.request.side_effect = [
            build_response(401, "/login-token", "nope"),
            build_response(401, "/login-token", "still no"),
            build_response(200, "/login-token", "fine."),
        ]
        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        client.ensure_auth()

        expected_call = mock.call("GET", "http://example.com/bodhi/oidc/login-token")
        self.oidc.request.call_args_list == [expected_call, expected_call, expected_call]
        self.oidc.clear_auth.call_count == 2
        self.oidc.login.call_count == 2

    def test_ensure_auth_failure(self):
        self.oidc.has_cookie.return_value = False
        self.oidc.request.return_value = build_response(500, "/login-token", "wat?")
        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        with pytest.raises(HTTPError):
            client.ensure_auth()

        self.oidc.request.assert_called_once_with(
            "GET", "http://example.com/bodhi/oidc/login-token"
        )
        self.oidc.login.assert_not_called()

    def test_send_request_ok(self, mocker):
        requests = mocker.patch("bodhi.client.bindings.requests")
        requests.request.return_value = build_response(200, "/url", '{"foo": "bar"}')

        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        response = client.send_request("somewhere", "GET")

        requests.request.assert_called_once_with("GET", "http://example.com/bodhi/somewhere")
        self.oidc.request.assert_not_called()
        assert "foo" in response
        assert response.foo == "bar"

    def test_send_request_with_auth(self, mocker):
        self.oidc.request.return_value = build_response(200, "/url", '{"foo": "bar"}')
        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        client.send_request("somewhere", "GET", auth=True)

        self.oidc.request.assert_called_once_with("GET", "http://example.com/bodhi/somewhere")
        self.oidc.ensure_auth.assert_called_once_with(use_kerberos=True)

    def test_send_request_error(self, mocker):
        response = build_response(500, "/url", "error")
        requests = mocker.patch("bodhi.client.bindings.requests")
        self.oidc.request.return_value = requests.request.return_value = response
        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        with pytest.raises(bindings.BodhiClientException) as exc:
            client.send_request("somewhere", "GET")
        assert str(exc.value) == "error"
        with pytest.raises(bindings.BodhiClientException) as exc:
            client.send_request("somewhere", "GET", auth=True)
        assert str(exc.value) == "error"

    def test_send_request_failure(self, mocker):
        failure = OIDCClientError("Something went wrong")
        requests = mocker.patch("bodhi.client.bindings.requests")
        self.oidc.request.side_effect = requests.request.side_effect = failure
        client = bindings.BodhiClient(base_url='http://example.com/bodhi/')

        with pytest.raises(bindings.BodhiClientException) as exc:
            client.send_request("somewhere", "GET")
        assert str(exc.value) == "Something went wrong"
        with pytest.raises(bindings.BodhiClientException) as exc:
            client.send_request("somewhere", "GET", auth=True)
        assert str(exc.value) == "Something went wrong"


class TestComment(BodhiClientTestCase):
    def test_comment(self, mocker):
        """
        Test the comment() method.
        """
        client = bindings.BodhiClient()
        client.csrf_token = 'a token'
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.comment('bodhi-2.4.0-1.fc25', 'It ate my cat!', karma=-1)

        assert response == 'response'
        client.send_request.assert_called_once_with(
            'comments/', verb='POST', auth=True,
            data={'update': 'bodhi-2.4.0-1.fc25', 'text': 'It ate my cat!', 'karma': -1,
                  'csrf_token': 'a token'})


class TestComposeStr(BodhiClientTestCase):
    def test_error_message(self, mocker):
        """Assert that an error message gets rendered in the long form."""
        mocker.patch.dict(client_test_data.EXAMPLE_COMPOSES_MUNCH['composes'][0],
                          {'error_message': 'some error'})
        s = bindings.BodhiClient.compose_str(
            client_test_data.EXAMPLE_COMPOSES_MUNCH['composes'][0], minimal=False)

        assert '*EPEL-7-stable  :   2 updates (requested)' in s
        assert 'Content Type: rpm' in s
        assert 'Started: 2018-03-15 17:25:22' in s
        assert 'Updated: 2018-03-15 17:25:22' in s
        assert 'Updates:' in s
        assert 'FEDORA-EPEL-2018-50566f0a39: uwsgi-2.0.16-1.el7' in s
        assert 'FEDORA-EPEL-2018-328e2b8c27: qtpass-1.2.1-3.el7' in s
        assert 'Error: some error' in s

    def test_minimal_false(self):
        """Test with minimal False."""
        s = bindings.BodhiClient.compose_str(
            client_test_data.EXAMPLE_COMPOSES_MUNCH['composes'][0], minimal=False)

        assert '*EPEL-7-stable  :   2 updates (requested)' in s
        assert 'Content Type: rpm' in s
        assert 'Started: 2018-03-15 17:25:22' in s
        assert 'Updated: 2018-03-15 17:25:22' in s
        assert 'Updates:' in s
        assert 'FEDORA-EPEL-2018-50566f0a39: uwsgi-2.0.16-1.el7' in s
        assert 'FEDORA-EPEL-2018-328e2b8c27: qtpass-1.2.1-3.el7' in s
        assert 'Error' not in s

    def test_minimal_true(self):
        """Test with minimal True."""
        s = bindings.BodhiClient.compose_str(
            client_test_data.EXAMPLE_COMPOSES_MUNCH['composes'][0], minimal=True)

        assert s == '*EPEL-7-stable  :   2 updates (requested) '

    def test_non_security_update(self, mocker):
        """Non-security updates should not have a leading *."""
        mocker.patch.dict(client_test_data.EXAMPLE_COMPOSES_MUNCH['composes'][0],
                          {'security': False})
        s = bindings.BodhiClient.compose_str(
            client_test_data.EXAMPLE_COMPOSES_MUNCH['composes'][0], minimal=True)

        assert s == ' EPEL-7-stable  :   2 updates (requested) '


class TestCSRF(BodhiClientTestCase):
    def test_with_csrf_token(self, mocker):
        """
        Test the method when csrf_token is set.
        """
        client = bindings.BodhiClient()
        client.csrf_token = 'a token'
        client.send_request = mocker.MagicMock(return_value='response')

        csrf = client.csrf()

        assert csrf == 'a token'
        assert client.send_request.call_count == 0

    def test_without_csrf_token_without_cookies(self, mocker):
        """
        Test the method when csrf_token is not set.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value={'csrf_token': 'a great token'})

        csrf = client.csrf()

        assert csrf == 'a great token'
        assert client.csrf_token == 'a great token'
        client.send_request.assert_called_once_with('csrf', verb='GET', auth=True)


class TestLatestBuilds(BodhiClientTestCase):
    def test_latest_builds(self, mocker):
        """
        Test latest_builds().
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='bodhi-2.4.0-1.fc25')

        latest_builds = client.latest_builds('bodhi')

        assert latest_builds == 'bodhi-2.4.0-1.fc25'
        client.send_request.assert_called_once_with('latest_builds', params={'package': 'bodhi'})


class TestCompose(BodhiClientTestCase):
    def test_get_compose_404_error(self, mocker):
        """
        Test for the case when the server returns a 404 error code.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        send_request.side_effect = HTTPError(
            response=build_response(
                404, "http://example.com/tests/composes/EPEL-7/stable", "update not found"
            )
        )

        with pytest.raises(bindings.ComposeNotFound) as exc:
            client.get_compose('EPEL-7', 'stable')

        assert exc.value.release == 'EPEL-7'
        assert exc.value.release_request == 'stable'

        send_request.assert_called_once_with('composes/EPEL-7/stable', verb='GET')

    def test_get_compose_successful_request(self, mocker):
        """
        Test with a successful request.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        send_request.return_value = client_test_data.EXAMPLE_COMPOSE_MUNCH

        response = client.get_compose('EPEL-7', 'stable')

        assert response == client_test_data.EXAMPLE_COMPOSE_MUNCH
        send_request.assert_called_once_with('composes/EPEL-7/stable', verb='GET')

    def test_get_compose_other_http_error(self, mocker):
        """
        Test for the case when a non-404 http error is raised.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        server_error = HTTPError(
            response=build_response(
                500, "http://example.com/tests/composes/EPEL-7/stable", "Internal server error"
            )
        )
        send_request.side_effect = server_error

        with pytest.raises(HTTPError) as exc:
            client.get_compose('EPEL-7', 'stable')

        assert exc.value is server_error

        send_request.assert_called_once_with('composes/EPEL-7/stable', verb='GET')

    def test_list_composes(self, mocker):
        """Assert a correct call to send_request() from list_composes()."""
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='some_composes')

        composes = client.list_composes()

        assert composes == 'some_composes'
        client.send_request.assert_called_once_with('composes/', verb='GET')


class TestListOverrides(BodhiClientTestCase):
    def test_with_user(self, mocker):
        """
        Test with the user parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(user='bowlofeggs')

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'user': 'bowlofeggs'})

    def test_without_parameters(self, mocker):
        """
        Test without the parameters.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides()

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET', params={})

    def test_with_package(self, mocker):
        """
        Test with the package parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(packages='bodhi')

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'packages': 'bodhi'})

    def test_with_expired(self, mocker):
        """
        Test --expired with the expired/active click boolean parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(expired=True)

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'expired': True})

    def test_with_active(self, mocker):
        """
        Test --active with the expired/active click boolean parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(expired=False)

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'expired': False})

    def test_with_releases(self, mocker):
        """
        Test with the releases parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(releases='F24')

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'releases': 'F24'})

    def test_list_overrides_with_builds(self, mocker):
        """
        Test with the builds parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(builds='python-1.5.6-3.fc26')

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'builds': 'python-1.5.6-3.fc26'})

    def test_list_overrides_with_rows_per_page(self, mocker):
        """
        Test with the rows_per_page parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(rows_per_page=10)

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'rows_per_page': 10})

    def test_list_overrides_with_page(self, mocker):
        """
        Test with the page parameter.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='response')

        response = client.list_overrides(page=5)

        assert response == 'response'
        client.send_request.assert_called_once_with('overrides/', verb='GET',
                                                    params={'page': 5})


class TestOverrideStr(BodhiClientTestCase):
    def test_with_min_dict(self):
        """
        Test override_str() with a dict argument and minimal set to true.
        """
        override = {
            'submitter': {'name': 'bowlofeggs'}, 'build': {'nvr': 'python-pyramid-1.5.6-3.el7'},
            'expiration_date': '2017-02-24'}

        override = bindings.BodhiClient.override_str(override)

        assert override == "bowlofeggs's python-pyramid-1.5.6-3.el7 override (expires 2017-02-24)"

    def test_with_dict(self):
        """
        Test override_str() with a dict argument.
        """
        override = {
            'submitter': {'name': 'bowlofeggs'}, 'build': {'nvr': 'js-tag-it-2.0-1.fc25'},
            'expiration_date': '2017-03-07 23:05:31', 'notes': 'No explanation given...',
            'expired_date': None}

        override = bindings.BodhiClient.override_str(override, minimal=False)

        assert override == client_test_data.EXPECTED_OVERRIDE_STR_OUTPUT.rstrip()

    def test_with_str(self):
        """
        Test override_str() with a str argument.
        """
        override = bindings.BodhiClient.override_str('this is an override')

        assert override == 'this is an override'


class TestQuery(BodhiClientTestCase):
    def test_with_bugs_empty_string(self, mocker):
        """
        Test with the bugs kwargs set to an empty string.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(builds='bodhi-2.4.0-1.fc26', bugs='')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'builds': 'bodhi-2.4.0-1.fc26', 'bugs': None})

    def test_with_limit(self, mocker):
        """
        Assert that the limit kwargs gets translated to rows_per_page correctly.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(builds='bodhi-2.4.0-1.fc26', limit=50)

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'builds': 'bodhi-2.4.0-1.fc26', 'rows_per_page': 50})

    def test_with_mine_false(self, mocker):
        """
        Assert correct behavior when the mine kwargs is False.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(builds='bodhi-2.4.0-1.fc26', mine=False)

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'builds': 'bodhi-2.4.0-1.fc26', 'mine': False})

    def test_with_mine_true(self, mocker):
        """
        Assert correct behavior when the mine kwargs is True.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.oidc.username = 'bowlofeggs'

        result = client.query(builds='bodhi-2.4.0-1.fc26', mine=True)

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET',
            params={'builds': 'bodhi-2.4.0-1.fc26', 'mine': True, 'user': 'bowlofeggs'})

    def test_with_mine_no_username(self, mocker):
        """
        Assert correct behavior when the mine kwargs is True but we are not authentified.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.oidc.username = None

        with pytest.raises(bindings.BodhiClientException) as exc:
            client.query(builds='bodhi-2.4.0-1.fc26', mine=True)
        assert str(exc.value) == "Could not get user info."

    def test_with_package_el_build(self, mocker):
        """
        Test with the package arg expressed as an el7 build.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(package='bodhi-2.4.0-1.el7')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'builds': 'bodhi-2.4.0-1.el7'})

    def test_with_package_epel_id(self, mocker):
        """
        Test with the package arg expressed as a EPEL update id.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(package='FEDORA-EPEL-2017-c3b112eb9e')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'updateid': 'FEDORA-EPEL-2017-c3b112eb9e'})

    def test_with_package_fc_build(self, mocker):
        """
        Test with the package arg expressed as a fc26 build.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(package='bodhi-2.4.0-1.fc26')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'builds': 'bodhi-2.4.0-1.fc26'})

    def test_with_package_fedora_id(self, mocker):
        """
        Test with the package arg expressed as a Fedora update id.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(package='FEDORA-2017-52506b30d4')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'updateid': 'FEDORA-2017-52506b30d4'})

    def test_with_package_name(self, mocker):
        """
        Test with the package arg expressed as a package name.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(package='bodhi')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'packages': 'bodhi'})

    def test_with_release_list(self, mocker):
        """
        Test with a 'release' kwarg set to a list.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(packages='bodhi', release=['f27'])

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'packages': 'bodhi', 'releases': ['f27']})

    def test_with_release_str(self, mocker):
        """
        Test with a 'release' kwarg set to a str.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(packages='bodhi', release='f26')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'packages': 'bodhi', 'releases': ['f26']})

    def test_with_type_(self, mocker):
        """
        Test with the type_ kwarg.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(packages='bodhi', type_='security')

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'packages': 'bodhi', 'type': 'security'})

    def test_query_with_rows_per_page(self, mocker):
        """
        Test with the 'rows_per_page' kwarg.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(packages='bodhi', rows_per_page=10)

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'packages': 'bodhi', 'rows_per_page': 10})

    def test_query_with_page(self, mocker):
        """
        Test with the 'page' kwarg.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')

        result = client.query(packages='bodhi', page=5)

        assert result == 'return_value'
        client.send_request.assert_called_once_with(
            'updates/', verb='GET', params={'packages': 'bodhi', 'page': 5})


class TestSave(BodhiClientTestCase):
    def test_save_with_type_(self, mocker):
        """
        Assert that save() handles type_ as a kwargs for backwards compatibility.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        kwargs = {
            'builds': ['bodhi-2.4.0-1.fc26'], 'type_': 'enhancement', 'notes': 'This is a test',
            'request': 'testing', 'autokarma': True, 'stable_karma': 3, 'unstable_karma': -3,
            'severity': 'low'}

        response = client.save(**kwargs)

        assert response == 'return_value'
        kwargs['type'] = kwargs['type_']
        kwargs['csrf_token'] = 'a token'
        client.send_request.assert_called_once_with('updates/', verb='POST', auth=True, data=kwargs)

    def test_without_type_(self, mocker):
        """
        Assert correct operation when type_ isn't given.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        kwargs = {
            'builds': ['bodhi-2.4.0-1.fc26'], 'type': 'enhancement', 'notes': 'This is a test',
            'request': 'testing', 'autokarma': True, 'stable_karma': 3, 'unstable_karma': -3,
            'severity': 'low'}

        response = client.save(**kwargs)

        assert response == 'return_value'
        kwargs['csrf_token'] = 'a token'
        client.send_request.assert_called_once_with('updates/', verb='POST', auth=True, data=kwargs)


class TestSaveOverride(BodhiClientTestCase):
    def test_save_override(self, mocker):
        """
        Test the save_override() method.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        now = datetime.utcnow()
        response = client.save_override(nvr='python-pyramid-1.5.6-3.el7',
                                        duration=2,
                                        notes='This is needed to build bodhi-2.4.0.')

        assert response == 'return_value'
        actual_expiration = client.send_request.mock_calls[0][2]['data']['expiration_date']
        client.send_request.assert_called_once_with(
            'overrides/', verb='POST', auth=True,
            data={'nvr': 'python-pyramid-1.5.6-3.el7',
                  'expiration_date': actual_expiration,
                  'csrf_token': 'a token', 'notes': 'This is needed to build bodhi-2.4.0.'})
        # Since we can't mock utcnow() since it's a C extension, let's just make sure the expiration
        # date sent is within 5 minutes of the now variable. It would be surprising if it took more
        # than 5 minutes to start the function and execute its first instruction!
        expected_expiration = now + timedelta(days=2)
        assert (actual_expiration - expected_expiration) < timedelta(minutes=5)

    def test_save_override_expiration_date(self, mocker):
        """
        Test the save_override() method with an explicit expiration date.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        now = datetime.utcnow()
        response = client.save_override(nvr='python-pyramid-1.5.6-3.el7',
                                        expiration_date=now,
                                        notes='This is needed to build bodhi-2.4.0.')

        assert response == 'return_value'
        client.send_request.assert_called_once_with(
            'overrides/', verb='POST', auth=True,
            data={'nvr': 'python-pyramid-1.5.6-3.el7',
                  'expiration_date': now,
                  'csrf_token': 'a token', 'notes': 'This is needed to build bodhi-2.4.0.'})

    def test_save_override_no_expiration(self, mocker):
        """
        Test the save_override() method without duration or expiration date.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        with pytest.raises(TypeError):
            client.save_override(
                nvr='python-pyramid-1.5.6-3.el7',
                notes='This is needed to build bodhi-2.4.0.'
            )

    def test_save_override_both_expirations(self, mocker):
        """
        Test the save_override() method with duration and expiration date.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        now = datetime.utcnow()
        with pytest.raises(TypeError):
            client.save_override(
                nvr='python-pyramid-1.5.6-3.el7',
                notes='This is needed to build bodhi-2.4.0.',
                duration=1,
                expiration_date=now,
            )

    def test_save_override_edit(self, mocker):
        """
        Test the save_override() method with the edit argument.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        now = datetime.utcnow()
        response = client.save_override(nvr='python-pyramid-1.5.6-3.el7',
                                        duration=2,
                                        notes='This is needed to build bodhi-2.4.0.',
                                        edit=True)

        assert response == 'return_value'
        actual_expiration = client.send_request.mock_calls[0][2]['data']['expiration_date']
        client.send_request.assert_called_once_with(
            'overrides/', verb='POST', auth=True,
            data={'nvr': 'python-pyramid-1.5.6-3.el7',
                  'expiration_date': actual_expiration,
                  'csrf_token': 'a token', 'notes': 'This is needed to build bodhi-2.4.0.',
                  'edited': 'python-pyramid-1.5.6-3.el7'})
        # Since we can't mock utcnow() since it's a C extension, let's just make sure the expiration
        # date sent is within 5 minutes of the now variable. It would be surprising if it took more
        # than 5 minutes to start the function and execute its first instruction!
        expected_expiration = now + timedelta(days=2)
        assert (actual_expiration - expected_expiration) < timedelta(minutes=5)

    def test_save_override_expired(self, mocker):
        """
        Test the save_override() method with the edit argument.
        """
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(return_value='return_value')
        client.csrf_token = 'a token'
        now = datetime.utcnow()
        response = client.save_override(nvr='python-pyramid-1.5.6-3.el7',
                                        duration=2,
                                        notes='This is needed to build bodhi-2.4.0.',
                                        expired=True)

        assert response == 'return_value'
        actual_expiration = client.send_request.mock_calls[0][2]['data']['expiration_date']
        client.send_request.assert_called_once_with(
            'overrides/', verb='POST', auth=True,
            data={'nvr': 'python-pyramid-1.5.6-3.el7',
                  'expiration_date': actual_expiration,
                  'csrf_token': 'a token', 'notes': 'This is needed to build bodhi-2.4.0.',
                  'expired': True})
        # Since we can't mock utcnow() since it's a C extension, let's just make sure the expiration
        # date sent is within 5 minutes of the now variable. It would be surprising if it took more
        # than 5 minutes to start the function and execute its first instruction!
        expected_expiration = now + timedelta(days=2)
        assert (actual_expiration - expected_expiration) < timedelta(minutes=5)


class TestRequest(BodhiClientTestCase):
    def test_request_404_error(self, mocker):
        """
        Test for the case when the server returns a 404 error code.
        """
        client = bindings.BodhiClient()
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        send_request.side_effect = HTTPError(
            response=build_response(
                404,
                "http://example.com/tests/updates/bodhi-2.2.4-99.el7/request",
                "update not found"
            )
        )

        with pytest.raises(bindings.UpdateNotFound) as exc:
            client.request('bodhi-2.2.4-1.el7', 'revoke')

        assert exc.value.update == 'bodhi-2.2.4-1.el7'

        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/request', verb='POST', auth=True,
            data={'csrf_token': 'a_csrf_token', 'request': 'revoke',
                  'update': 'bodhi-2.2.4-1.el7'})

    def test_request_successful_request(self, mocker):
        """
        Test with a successful request.
        """
        client = bindings.BodhiClient()
        send_request = mocker.patch.object(client, "send_request")
        send_request.return_value = client_test_data.EXAMPLE_UPDATE_MUNCH
        client.csrf_token = 'a_csrf_token'

        response = client.request('bodhi-2.2.4-1.el7', 'revoke')

        assert response == client_test_data.EXAMPLE_UPDATE_MUNCH
        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/request', verb='POST', auth=True,
            data={'csrf_token': 'a_csrf_token', 'request': 'revoke',
                  'update': 'bodhi-2.2.4-1.el7'})

    def test_request_other_http_error(self, mocker):
        """
        Test for the case when a non-404 http error is raised.
        """
        client = bindings.BodhiClient()
        send_request = mocker.patch.object(client, "send_request")
        client.csrf_token = 'a_csrf_token'
        server_error = HTTPError(
            response=build_response(
                500,
                "http://example.com/tests/updates/bodhi-2.2.4-99.el7/request",
                "Internal server error",
            )
        )
        send_request.side_effect = server_error

        with pytest.raises(HTTPError) as exc:
            client.request('bodhi-2.2.4-1.el7', 'revoke')

        assert exc.value is server_error

        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/request', verb='POST', auth=True,
            data={'csrf_token': 'a_csrf_token', 'request': 'revoke',
                  'update': 'bodhi-2.2.4-1.el7'})


class TestUpdateStr(BodhiClientTestCase):
    def test_bugs(self, mocker):
        """Ensure correct output when there are bugs on the update."""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'bugs': [{'bug_id': 1234, 'title': 'it broke'}, {'bug_id': 1235, 'title': 'halp'}]}
        )
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert compare_output(
            text,
            client_test_data.EXPECTED_UPDATE_OUTPUT.replace(
                'Autotime: True', 'Autotime: True\n   Bugs: 1234 - it broke\n    : 1235 - halp'))

    def test_minimal(self, mocker):
        """Ensure correct output when minimal is True."""
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        client = bindings.BodhiClient()
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 24, 12, 0, 0))
        mock_datetime.strptime = datetime.strptime

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = (' bodhi-2.2.4-1.el7                        rpm        stable    '
                           '2016-10-21 (2)')
        assert text == expected_output

    def test_minimal_not_pushed(self, mocker):
        """Ensure correct output when minimal is True and not yet pushed."""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'date_pushed': '', 'pushed': False, 'status': 'pending'}
        )
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        client = bindings.BodhiClient()
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 5, 23, 0, 0))
        mock_datetime.strptime = datetime.strptime

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = (' bodhi-2.2.4-1.el7                        rpm        pending   '
                           '2016-10-05 (0)')
        assert text == expected_output

    def test_minimal_type_security(self, mocker):
        """Ensure correct output when minimal is True and type security"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'type': 'security'}
        )
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        client = bindings.BodhiClient()
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 24, 12, 0, 0))
        mock_datetime.strptime = datetime.strptime

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = ('*bodhi-2.2.4-1.el7                        rpm        stable    '
                           '2016-10-21 (2)')
        assert text == expected_output

    def test_minimal_with_multiple_builds(self, mocker):
        """Ensure correct output when minimal is True, and multiple builds"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'builds': [
                {'epoch': 0, 'nvr': 'bodhi-2.2.4-1.el7', 'signed': True},
                {'epoch': 0, 'nvr': 'bodhi-pants-2.2.4-1.el7', 'signed': True}
            ]}
        )
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 24, 12, 0, 0))
        mock_datetime.strptime = datetime.strptime
        client = bindings.BodhiClient()

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = (' bodhi-2.2.4-1.el7                        rpm        stable    '
                           '2016-10-21 (2)\n  bodhi-pants-2.2.4-1.el7')
        assert text == expected_output

    def test_minimal_no_builds(self, mocker):
        """Ensure correct output when minimal is True, and there are no builds"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'builds': [], 'title': 'update-title'}
        )
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        client = bindings.BodhiClient()
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 24, 12, 0, 0))
        mock_datetime.strptime = datetime.strptime

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = (' update-title                             rpm        stable    '
                           '2016-10-21 (2)')
        assert text == expected_output

    def test_minimal_no_title(self, mocker):
        """Ensure correct output when minimal is True, and there are neither a title nor builds"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'title': None, 'builds': []}
        )
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 24, 12, 0, 0))
        mock_datetime.strptime = datetime.strptime
        client = bindings.BodhiClient()

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = (' FEDORA-EPEL-2016-3081a94111              '
                           'rpm        stable    '
                           '2016-10-21 (2)')
        assert text == expected_output

    def test_minimal_no_content_type(self, mocker):
        """Ensure correct output when minimal is True, and and there is no content-type"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'content_type': None}
        )
        mock_datetime = mocker.patch("bodhi.client.bindings.datetime.datetime")
        mock_datetime.utcnow = mock.Mock(return_value=datetime(2016, 10, 24, 12, 0, 0))
        mock_datetime.strptime = datetime.strptime
        client = bindings.BodhiClient()

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH, minimal=True)

        expected_output = (' bodhi-2.2.4-1.el7                        '
                           'unspecified  stable    '
                           '2016-10-21 (2)')
        assert text == expected_output

    def test_request_stable(self, mocker):
        """Ensure correct output when the update is request stable."""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'request': 'stable'}
        )
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)
        print("This is", text)

        assert compare_output(
            text,
            client_test_data.EXPECTED_UPDATE_OUTPUT.replace(
                'Autotime: True', 'Autotime: True\n     Request: stable'))

    def test_severity(self, mocker):
        """Test that severity is rendered."""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'Severity: unspecified' in text

    def test_with_autokarma_set(self, mocker):
        """
        Ensure correct operation when autokarma is True..
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert compare_output(text, client_test_data.EXPECTED_UPDATE_OUTPUT)

    def test_with_autokarma_unset(self, mocker):
        """
        Ensure correct operation when autokarma is False.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")
        update = copy.deepcopy(client_test_data.EXAMPLE_UPDATE_MUNCH)
        # Set the update's autokarma and to False.
        update.autokarma = False

        text = client.update_str(update)

        expected_output = client_test_data.EXPECTED_UPDATE_OUTPUT.replace(
            'Autokarma: True  [-3, 3]', 'Autokarma: False  [-3, 3]')
        assert compare_output(text, expected_output)

    def test_autotime_set(self, mocker):
        """
        Ensure correct operation when autotime is True.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert compare_output(text, client_test_data.EXPECTED_UPDATE_OUTPUT)

    def test_autotime_unset(self, mocker):
        """
        Ensure correct operation when autotime is False.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")
        update = copy.deepcopy(client_test_data.EXAMPLE_UPDATE_MUNCH)
        # Set the update's autotime and to False.
        update.autotime = False

        text = client.update_str(update)

        expected_output = client_test_data.EXPECTED_UPDATE_OUTPUT.replace(
            'Autotime: True', 'Autotime: False')
        assert compare_output(text, expected_output)

    def test_update_as_string(self):
        """Ensure we return a string if update is a string"""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'

        text = client.update_str("this is a string")

        assert text == "this is a string"

    def test_update_with_unicode_comment(self, mocker):
        """Ensure unicode content in update comments is correctly handled"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH.comments[0],
            {'text': 'This comment contains a unicode char ☺. '}
        )
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert compare_output(
            text,
            client_test_data.EXPECTED_UPDATE_OUTPUT.replace(
                'This update has been submitted for testing by bowlofeggs.',
                'This comment contains a unicode char ☺.'))

    def test_update_with_unicode_note(self, mocker):
        """Ensure unicode content in update notes is correctly handled"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'notes': 'This note contains a unicode char ☺'}
        )
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'Notes: This note contains a unicode char ☺' in text

    def test_ci_status_errors(self, mocker):
        """Ensure that ci error is displayed"""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        get_test_status = mocker.patch.object(client, "get_test_status")
        get_test_status.return_value = munch.Munch(
            {'errors': [munch.Munch({'description': 'bar'})]})

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'CI Status: bar\n' in text

    def test_ci_status_failure(self, mocker):
        """Ensure that ci is not displayed when it fails"""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        get_test_status = mocker.patch.object(client, "get_test_status")
        get_test_status.side_effect = HTTPError("testing error")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'CI Status' not in text

    def test_ci_status(self, mocker):
        """Ensure that ci information is displayed"""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        get_test_status = mocker.patch.object(client, "get_test_status")
        get_test_status.return_value = munch.Munch(
            {'decision': munch.Munch({'summary': 'no tests required', 'waivers': []})}
        )

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'CI Status: no tests required\n' in text

    def test_waived_tests(self, mocker):
        """Ensure that information about waived tests is rendered"""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        get_test_status = mocker.patch.object(client, "get_test_status")
        get_test_status.return_value = munch.Munch(
            {'decision': munch.Munch({
                'summary': 'no tests required',
                'waivers': [{'comment': 'This is fine. See BZ#1566485', 'id': 150,
                             'product_version': 'fedora-28', 'proxied_by': None,
                             'subject': {'item': 'slop-7.4-1.fc28', 'type': 'koji_build'},
                             'subject_identifier': 'slop-7.4-1.fc28', 'subject_type': 'koji_build',
                             'testcase': 'dist.rpmlint', 'timestamp': '2018-06-29T00:20:20.425844',
                             'username': 'netvor', 'waived': True}]})}
        )

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert (
            '     Waivers: netvor - 2018-06-29 00:20:20\n'
            '              This is fine. See BZ#1566485\n'
            '              build: slop-7.4-1.fc28\n'
            '              testcase: dist.rpmlint\n'
            in text
        )

    def test_ci_status_new_format(self, mocker):
        """Ensure that ci information is displayed with Greenwave's new format"""
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        get_test_status = mocker.patch.object(client, "get_test_status")
        get_test_status.return_value = munch.Munch(
            {'decisions': [munch.Munch({'summary': 'no tests required', 'waivers': []})]}
        )

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'CI Status: no tests required\n' in text

    def test_notes_multiline(self, mocker):
        """Ensure that multiline notes are rendered"""
        mocker.patch.dict(
            client_test_data.EXAMPLE_UPDATE_MUNCH,
            {'notes': 'This note contains:\n* multiline formatting\n* bullet points\n\n'}
        )
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        mocker.patch.object(client, "send_request")

        text = client.update_str(client_test_data.EXAMPLE_UPDATE_MUNCH)

        assert 'Notes: This note contains:\n' in text
        assert '     : * multiline formatting\n' in text
        assert '     : * bullet points\n' in text


class TestErrorHandled:
    def test_failure_with_given_errors(self):
        """
        Test the failure case for when errors were given in the response.
        """
        @bindings.errorhandled
        def im_gonna_fail_but_ill_be_cool_about_it(x, y, z=None):
            assert x == 1
            assert y == 2
            assert z == 3

            return {'errors': [{'description': 'insert'}, {'description': 'coin(s)'}]}

        with pytest.raises(bindings.BodhiClientException) as exc:
            im_gonna_fail_but_ill_be_cool_about_it(1, y=2, z=3)

        assert str(exc.value) == 'insert\ncoin(s)'

    def test_retry_on_captcha_key_failure(self, mocker):
        """
        Test the decorator when the wrapped method returns a captch_key error.

        This test ensures that the decorator will retry the wrapped method if it returns a
        captcha_key error, after clearing cookies and the csrf token.

        This test was written to assert the fix for
        https://github.com/fedora-infra/bodhi/issues/1787
        """
        a_fake_self = mocker.MagicMock()
        a_fake_self.csrf_token = 'some_token'
        a_fake_self.call_count = 0

        @bindings.errorhandled
        def captcha_plz(a_fake_self):
            a_fake_self.call_count = a_fake_self.call_count + 1

            # Fail on the first call with a captcha_key error to simulate unauth'd user on a
            # comment.
            if a_fake_self.call_count == 1:
                return {'errors': [{'name': 'captcha_key'}]}

            return 'here you go'

        # No Exception should be raised.
        captcha_plz(a_fake_self)

        a_fake_self.clear_auth.assert_called_once_with()
        assert a_fake_self.call_count == 2

    def test_success(self):
        """
        Test the decorator for the success case.
        """
        @bindings.errorhandled
        def im_gonna_be_cool(x, y, z=None):
            assert x == 1
            assert y == 2
            assert z == 3

            return 'here you go'

        assert im_gonna_be_cool(1, 2, 3) == 'here you go'

    def test_unexpected_error(self):
        """
        Test the failure case when errors are not given in the response.
        """
        @bindings.errorhandled
        def im_gonna_fail_and_i_wont_be_cool_about_it(x, y, z=None):
            assert x == 1
            assert y == 2
            assert z == 3

            return {'errors': ['MEAN ERROR']}

        with pytest.raises(bindings.BodhiClientException) as exc:
            im_gonna_fail_and_i_wont_be_cool_about_it(1, 2, z=3)

        assert str(exc.value) == 'An unhandled error occurred in the BodhiClient'


class TestUpdateNotFound:
    def test_updatenotfound(self):
        """
        Assert that __init__() works properly.
        """
        exc = bindings.UpdateNotFound('bodhi-2.2.4-1.el7')

        assert exc.update == 'bodhi-2.2.4-1.el7'
        assert isinstance(exc.update, str)

    def test_updatenotfound_str(self):
        """
        Assert that __str__() works properly.
        """
        exc = bindings.UpdateNotFound('bodhi-2.2.4-1.el7')

        assert str(exc.update) == 'bodhi-2.2.4-1.el7'
        assert isinstance(str(exc.update), str)
        assert str(exc) == 'Update not found: bodhi-2.2.4-1.el7'


class TestCandidates(BodhiClientTestCase):
    def test_candidates_failure(self, mocker, caplog):
        """Ensure correct handling when talking to Koji raises an Exception."""
        client = bindings.BodhiClient()
        client.oidc.username = 'bowlofeggs'
        get_koji_session = mocker.patch.object(client, "get_koji_session")
        get_koji_session.return_value.listTagged.side_effect = [
            [
                {'name': 'bodhi', 'version': '2.9.0', 'release': '1.fc25',
                 'nvr': 'bodhi-2.9.0-1.fc25', 'owner_name': 'bowlofeggs'},
                {'name': 'ipsilon', 'version': '2.0.2', 'release': '1.fc25',
                 'nvr': 'ipsilon-2.0.2-1.fc25', 'owner_name': 'puiterwijk'}
            ],
            IOError("Bet you didn't expect this.")
        ]
        mocker.patch.object(client, "send_request", return_value={
            'releases': [
                {'candidate_tag': 'f25-updates-testing'},
                {'candidate_tag': 'f26-updates-testing'}
            ]
        })

        results = client.candidates()

        assert results == [{'release': '1.fc25',
                            'version': '2.9.0',
                            'name': 'bodhi',
                            'owner_name': 'bowlofeggs',
                            'nvr': 'bodhi-2.9.0-1.fc25'}]
        get_koji_session.assert_called_once_with()
        assert (
            get_koji_session.return_value.listTagged.mock_calls
            == [mock.call('f25-updates-testing', latest=True),
                mock.call('f26-updates-testing', latest=True)])
        client.send_request.assert_called_once_with('releases/', params={}, verb='GET')
        expected_error = (
            "Unable to query candidate builds for {'candidate_tag': 'f26-updates-testing'}"
        )
        assert expected_error in caplog.messages

    def test_candidates_success(self, mocker):
        """Ensure correct behavior when there are no errors talking to Koji."""
        client = bindings.BodhiClient()
        client.oidc.username = 'bowlofeggs'
        get_koji_session = mocker.patch.object(client, "get_koji_session")
        get_koji_session.return_value.listTagged.side_effect = [
            [{'name': 'bodhi', 'version': '2.9.0', 'release': '1.fc25', 'nvr': 'bodhi-2.9.0-1.fc25',
              'owner_name': 'bowlofeggs'},
             {'name': 'ipsilon', 'version': '2.0.2', 'release': '1.fc25',
              'nvr': 'ipsilon-2.0.2-1.fc25', 'owner_name': 'puiterwijk'}],
            [{'name': 'bodhi', 'version': '2.9.0', 'release': '1.fc26', 'nvr': 'bodhi-2.9.0-1.fc26',
              'owner_name': 'bowlofeggs'}]]
        mocker.patch.object(client, "send_request", return_value={
            'releases': [
                {'candidate_tag': 'f25-updates-testing'},
                {'candidate_tag': 'f26-updates-testing'}
            ]
        })

        results = client.candidates()

        assert results == [
            {
                'release': '1.fc25',
                'version': '2.9.0',
                'name': 'bodhi',
                'owner_name': 'bowlofeggs',
                'nvr': 'bodhi-2.9.0-1.fc25'
            },
            {
                'release': '1.fc26',
                'version': '2.9.0',
                'name': 'bodhi',
                'owner_name': 'bowlofeggs',
                'nvr': 'bodhi-2.9.0-1.fc26'
            }
        ]
        get_koji_session.assert_called_once_with()
        assert (
            get_koji_session.return_value.listTagged.mock_calls
            == [mock.call('f25-updates-testing', latest=True),
                mock.call('f26-updates-testing', latest=True)])
        client.send_request.assert_called_once_with('releases/', params={}, verb='GET')


class TestGetReleases(BodhiClientTestCase):
    def test_get_releases(self, mocker):
        """Assert correct behavior from the get_releases() method."""
        client = bindings.BodhiClient()
        client.send_request = mocker.MagicMock(
            return_value={'releases': [{'candidate_tag': 'f25-updates-testing'},
                                       {'candidate_tag': 'f26-updates-testing'}]})

        results = client.get_releases(some_param='some_value')

        assert results == {
            'releases': [
                {'candidate_tag': 'f25-updates-testing'},
                {'candidate_tag': 'f26-updates-testing'}
            ]
        }
        client.send_request.assert_called_once_with(
            'releases/', params={'some_param': 'some_value'}, verb='GET')


class TestParseFile(BodhiClientTestCase):
    def test_parsing_invalid_file(self, tmpdir, mocker):
        """
        Test parsing an invalid update template file.
        """
        filepath = tmpdir.join("f.ini")
        open(filepath, "w").close()
        cp_read = mocker.patch("bodhi.client.bindings.configparser.ConfigParser.read")
        # This happens when we don't have permission to read the file.
        # We can't just remove the permissions because the unit tests run as root in bodhi-ci.
        cp_read.return_value = []
        client = bindings.BodhiClient()

        with pytest.raises(ValueError) as exc:
            try:
                result = client.parse_file(filepath)
            except Exception as e:
                print(e)
                raise
            print(result)

        assert str(exc.value) == f'Invalid input file: {filepath}'

    def test_parsing_valid_file(self, tmpdir):
        """
        Test parsing a valid update template file
        """
        s = [
            "[fedora-workstation-backgrounds-1.1-1.fc26]\n",
            "# bugfix, security, enhancement, newpackage (required)\n",
            "type=bugfix\n",
            "\n",
            "# testing, stable\n",
            "request=testing\n",
            "\n",
            "# Bug numbers: 1234,9876\n",
            "bugs=123456,43212\n",
            "\n",
            "# Here is where you give an explanation of your update.\n",
            "notes=Initial Release\n",
            "\n",
            "# Update name\n",
            "display_name=fake update name\n",
            "\n",
            "# Enable request automation based on the stable/unstable karma thresholds\n",
            "autokarma=True\n",
            "stable_karma=3\n",
            "unstable_karma=-3\n",
            "\n",
            "# Automatically close bugs when this marked as stable\n",
            "close_bugs=True\n",
            "\n",
            "# Suggest that users restart after update\n",
            "suggest_reboot=False\n",
            ""]

        filepath = tmpdir.join("f.ini")
        with open(filepath, "w") as f:
            f.writelines(s)
        client = bindings.BodhiClient()
        updates = client.parse_file(filepath)

        assert len(updates) == 1
        assert len(updates[0]) == 13
        assert updates[0]['close_bugs'] is True
        assert updates[0]['display_name'] == 'fake update name'
        assert updates[0]['unstable_karma'] == '-3'
        assert updates[0]['severity'] == 'unspecified'
        assert updates[0]['stable_karma'] == '3'
        assert updates[0]['builds'] == 'fedora-workstation-backgrounds-1.1-1.fc26'
        assert updates[0]['autokarma'] == 'True'
        assert updates[0]['suggest'] == 'unspecified'
        assert updates[0]['notes'] == 'Initial Release'
        assert updates[0]['request'] == 'testing'
        assert updates[0]['bugs'] == '123456,43212'
        assert updates[0]['type_'] == 'bugfix'
        assert updates[0]['type'] == 'bugfix'

    def test_parsing_nonexistent_file(self):
        """
        Test trying to parse a file that doesnt exist
        """
        client = bindings.BodhiClient()

        with pytest.raises(ValueError) as exc:
            client.parse_file("/tmp/bodhi-test-parsefile2")

        assert str(exc.value) == 'No such file or directory: /tmp/bodhi-test-parsefile2'


class TestTestable(BodhiClientTestCase):
    def test_testable(self, mocker):
        """Assert correct behavior from the testable() method."""
        dnf = mocker.patch("bodhi.client.bindings.dnf")
        fill_sack = mock.MagicMock()
        dnf.Base.return_value.fill_sack = fill_sack
        fill_sack.return_value.query.return_value.installed.return_value.filter.\
            return_value.run.return_value = ['bodhi-2.8.1-1.fc26']

        client = bindings.BodhiClient()
        get_koji_session = mocker.patch.object(client, "get_koji_session")
        get_koji_session.return_value.listTagged.return_value = [
            {'name': 'bodhi', 'version': '2.9.0', 'release': '1.fc26', 'nvr': 'bodhi-2.9.0-1.fc26'}]

        client.send_request = mock.MagicMock(
            return_value={'updates': [{'nvr': 'bodhi-2.9.0-1.fc26'}]})

        mock_open = mocker.patch('builtins.open', create=True)
        mock_open.return_value.__enter__.return_value.readlines.return_value = [
            'Fedora release 26 (Twenty Six)']

        updates = client.testable()

        assert list(updates) == [{'nvr': 'bodhi-2.9.0-1.fc26'}]
        fill_sack.assert_called_once_with(load_system_repo=True)
        fill_sack.return_value.query.assert_called_once_with()
        fill_sack.return_value.query.return_value.installed.assert_called_once_with()
        fill_sack.return_value.query.return_value.installed.return_value.filter.\
            assert_called_once_with(name='bodhi', version='2.9.0', release='1.fc26')
        fill_sack.return_value.query.return_value.installed.return_value.filter.return_value.run.\
            assert_called_once_with()
        get_koji_session.return_value.listTagged.assert_called_once_with('f26-updates-testing',
                                                                         latest=True)
        client.send_request.assert_called_once_with(
            'updates/', params={'builds': 'bodhi-2.9.0-1.fc26'}, verb='GET')

    def test_testable_no_dnf(self, mocker):
        """Ensure that testable raises a RuntimeError if dnf is None."""
        mocker.patch("bodhi.client.bindings.dnf", None)

        client = bindings.BodhiClient()

        with pytest.raises(RuntimeError) as exc:
            list(client.testable())

        assert str(exc.value) == 'dnf is required by this method and is not installed.'


class TestKojiSession(BodhiClientTestCase):
    def test_koji_conf_in_home_directory(self, mocker, tmpdir):
        """Test that if ~/.koji/config exists, we read the config from there first"""
        koji = mocker.patch("bodhi.client.bindings.koji.ClientSession")
        mocker.patch("os.path.expanduser", return_value=tmpdir)
        os.mkdir(tmpdir.join(".koji"))
        with open(tmpdir.join(".koji", "config"), "w") as f:
            f.write("[koji]\nserver = foobar\n")
        client = bindings.BodhiClient()
        client.get_koji_session()
        koji.assert_called_with("foobar")

    def test_koji_conf_not_in_home_directory(self, mocker, tmpdir):
        """Test that if ~/.koji/config doesn't exist, we read config from /etc/koji.conf"""
        koji = mocker.patch("bodhi.client.bindings.koji.ClientSession")
        with open(tmpdir.join("etckojiconf"), "w") as f:
            f.write("[koji]\nserver = foobar\n")

        client = bindings.BodhiClient()
        with open(tmpdir.join("etckojiconf")) as f:
            mock_open = mocker.patch('builtins.open', create=True)
            mock_open.return_value = f
            # read_file = mocker.patch('bodhi.client.bindings.configparser.ConfigParser.read_file')
            client.get_koji_session()
            mock_open.assert_called_once_with("/etc/koji.conf")
        koji.assert_called_with("foobar")


class TestWaive(BodhiClientTestCase):
    def test_waive_404_error(self, mocker):
        """
        Test for the case when the server returns a 404 error code.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        server_error = HTTPError(
            response=build_response(
                404,
                "http://example.com/tests/updates/bodhi-2.2.4-99.el7/waive-test-results",
                "update not found",
            )
        )
        send_request.side_effect = server_error

        with pytest.raises(bindings.UpdateNotFound) as exc:
            client.waive('bodhi-2.2.4-1.el7', comment='Expected failure', tests=None)

        assert exc.value.update == 'bodhi-2.2.4-1.el7'

        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/waive-test-results', verb='POST', auth=True,
            data={'comment': 'Expected failure', 'csrf_token': 'a_csrf_token',
                  'tests': None, 'update': 'bodhi-2.2.4-1.el7'})

    def test_successful_waive_some(self, mocker):
        """
        Test with a successful request.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        send_request.return_value = client_test_data.EXAMPLE_UPDATE_MUNCH

        response = client.waive(
            'bodhi-2.2.4-1.el7', comment='Expected failure',
            tests=('dist.rpmdeplint', 'fedora-atomic-ci')
        )

        assert response == client_test_data.EXAMPLE_UPDATE_MUNCH
        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/waive-test-results', verb='POST', auth=True,
            data={'comment': 'Expected failure', 'csrf_token': 'a_csrf_token',
                  'tests': ('dist.rpmdeplint', 'fedora-atomic-ci'), 'update': 'bodhi-2.2.4-1.el7'})

    def test_successful_waive_all(self, mocker):
        """
        Test with a successful request.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        send_request.return_value = client_test_data.EXAMPLE_UPDATE_MUNCH

        response = client.waive('bodhi-2.2.4-1.el7', comment='Expected failure', tests=None)

        assert response == client_test_data.EXAMPLE_UPDATE_MUNCH
        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/waive-test-results', verb='POST', auth=True,
            data={'comment': 'Expected failure', 'csrf_token': 'a_csrf_token',
                  'tests': None, 'update': 'bodhi-2.2.4-1.el7'})

    def test_waive_other_http_error(self, mocker):
        """
        Test for the case when a non-404 http error is raised.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        server_error = HTTPError(
            response=build_response(
                500,
                "http://example.com/tests/updates/bodhi-2.2.4-99.el7/waive-test-results",
                "Internal server error",
            )
        )
        send_request.side_effect = server_error

        with pytest.raises(HTTPError) as exc:
            client.waive('bodhi-2.2.4-1.el7', comment='Expected failure', tests=None)

        assert exc.value is server_error

        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/waive-test-results', verb='POST', auth=True,
            data={'comment': 'Expected failure', 'csrf_token': 'a_csrf_token',
                  'tests': None, 'update': 'bodhi-2.2.4-1.el7'})


class TestTriggerTests(BodhiClientTestCase):
    def test_trigger_tests_404_error(self, mocker):
        """
        Test for the case when the server returns a 404 error code.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        server_error = HTTPError(
            response=build_response(
                404,
                "http://example.com/tests/updates/bodhi-2.2.4-99.el7/trigger-tests",
                "update not found",
            )
        )
        send_request.side_effect = server_error

        with pytest.raises(bindings.UpdateNotFound) as exc:
            client.trigger_tests('bodhi-2.2.4-1.el7')

        assert exc.value.update == 'bodhi-2.2.4-1.el7'

        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/trigger-tests', verb='POST', auth=True,
            data={'csrf_token': 'a_csrf_token',
                  'update': 'bodhi-2.2.4-1.el7'})

    def test_successful_trigger(self, mocker):
        """
        Test with a successful request.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        send_request.return_value = client_test_data.EXAMPLE_UPDATE_MUNCH

        response = client.trigger_tests(
            'bodhi-2.2.4-1.el7')

        assert response == client_test_data.EXAMPLE_UPDATE_MUNCH
        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/trigger-tests', verb='POST', auth=True,
            data={'csrf_token': 'a_csrf_token',
                  'update': 'bodhi-2.2.4-1.el7'})

    def test_trigger_tests_other_http_error(self, mocker):
        """
        Test for the case when a non-404 http error is raised.
        """
        client = bindings.BodhiClient()
        client.base_url = 'http://example.com/tests/'
        client.csrf_token = 'a_csrf_token'
        send_request = mocker.patch.object(client, "send_request")
        server_error = HTTPError(
            response=build_response(
                500,
                "http://example.com/tests/updates/bodhi-2.2.4-99.el7/trigger-tests",
                "Internal server error",
            )
        )
        send_request.side_effect = server_error

        with pytest.raises(HTTPError) as exc:
            client.trigger_tests('bodhi-2.2.4-1.el7')

        assert exc.value is server_error

        send_request.assert_called_once_with(
            'updates/bodhi-2.2.4-1.el7/trigger-tests', verb='POST', auth=True,
            data={'csrf_token': 'a_csrf_token',
                  'update': 'bodhi-2.2.4-1.el7'})



# Copyright © 2014-2019 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The bodhi CLI client."""

from datetime import datetime
import functools
import logging
import os
import platform
import re
import subprocess
import sys
import traceback
import typing

import click
import munch

from bodhi.client import bindings, constants


log = logging.getLogger(__name__)


def _warn_staging_overrides(
        ctx: click.core.Context, param: typing.Union[click.core.Option, click.core.Parameter],
        value: str) -> str:
    """
    Print a warning to stderr if the user has set both the --url/--id-provider and --staging flags.

    This ensures that the user is aware that --staging supersedes --url/--id-provider.

    Args:
        ctx: The Click context, used to find out if the --staging flag is set.
        param: The option being handled.
        value: The value of the option being handled.
    Returns:
        The value of the option being handled.
    """
    if ctx.params.get('staging', False):
        if (
            param.name == 'url' and value != constants.BASE_URL
            or param.name == 'id_provider' and value != constants.IDP
        ):
            click.echo(
                f'\nWarning: {param.name} and staging flags are '
                f'both set. {param.name} will be ignored.\n',
                err=True
            )
    if param.name == 'staging' and value:
        if ctx.params.get('url', constants.BASE_URL) != constants.BASE_URL:
            click.echo('\nWarning: url and staging flags are both set. url will be ignored.\n',
                       err=True)
        if ctx.params.get('id_provider', constants.IDP) != constants.IDP:
            click.echo('\nWarning: id_provider and staging flags '
                       'are both set. id_provider will be ignored.\n',
                       err=True)
    return value


def _set_logging_debug(
        ctx: click.core.Context, param: typing.Union[click.core.Option, click.core.Parameter],
        value: bool) -> bool:
    """
    Set up the logging level to "debug".

    This allows us to print more information on the user's screen and thus helps following what is
    going on.

    Args:
        ctx: The Click context. Unused.
        param: The option being handled. Unused.
        value: The value of the --debug flag.
    Returns:
        The value of the --debug flag.
    """
    if value:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        root_logger = logging.getLogger("root")
        root_logger.addHandler(ch)
        root_logger.setLevel(logging.DEBUG)
    return value


url_option = click.option('--url', envvar='BODHI_URL', default=constants.BASE_URL,
                          help=('URL of a Bodhi server. Ignored if --staging is set. Can be set '
                                'with BODHI_URL environment variable'),
                          callback=_warn_staging_overrides)
openid_options = [
    click.option(
        "--client-id", default=constants.CLIENT_ID,
        help="The OIDC client_id. Ignored if --staging is set."),
    click.option(
        "--id-provider", default=constants.IDP, envvar='BODHI_OPENID_PROVIDER',
        callback=_warn_staging_overrides,
        help="The OIDC provider. Ignored if --staging is set.")
]

staging_option = click.option('--staging', help='Use the staging bodhi instance',
                              is_flag=True, default=False,
                              callback=_warn_staging_overrides)
debug_option = click.option('--debug', help='Display debugging information.',
                            is_flag=True, default=False,
                            callback=_set_logging_debug)


new_edit_options = [
    click.option('--severity', help='Update severity',
                 type=click.Choice(['unspecified', 'low', 'medium', 'high', 'urgent']),
                 is_eager=True),
    click.option('--notes', help='Update description'),
    click.option('--notes-file', help='Update description from a file'),
    click.option('--bugs', help='Comma-separated list of bug numbers', default=''),
    click.option('--close-bugs/--no-close-bugs', default=None,
                 help='Automatically close bugs or not'),
    click.option('--request', help='Requested repository',
                 type=click.Choice(constants.REQUEST_TYPES)),
    click.option('--autotime/--no-autotime', default=None,
                 help='Enable/Disable stable push based on time in testing'),
    click.option('--stable-days', type=click.INT,
                 help='Days in testing required to push to stable'),
    click.option('--autokarma/--no-autokarma', default=None,
                 help='Enable/Disable karma automatism'),
    click.option('--stable-karma', type=click.INT, help='Stable karma threshold'),
    click.option('--unstable-karma', type=click.INT, help='Unstable karma threshold'),
    click.option('--requirements',
                 help='Space or comma-separated list of required Taskotron tasks'),
    click.option('--suggest', help='Post-update user suggestion',
                 type=click.Choice(constants.SUGGEST_TYPES)),
    click.option('--display-name',
                 help='The name of the update', default=None),
    staging_option]


# Common options for the overrides save and edit command
save_edit_options = [
    click.argument('nvr'),
    click.option('--duration', default=7, type=click.INT,
                 help='Number of days the override should exist.'),
    click.option('--notes', default="No explanation given...",
                 help='Notes on why this override is in place.'),
    click.option('--wait/--no-wait', is_flag=True, default=True,
                 help='Wait and ensure that the override is active'),
    staging_option,
    url_option,
    debug_option]


# Basic options for pagination of query result
pagination_options = [
    click.option('--rows', default=None,
                 type=click.IntRange(1, 100, clamp=False),
                 help='Limits number of results shown per page'),
    click.option('--page', default=None,
                 type=click.IntRange(1, clamp=False),
                 help='Go to page number')]


# Common releases options
release_options = [
    click.option('--name', help='Release name (eg: F20)'),
    click.option('--long-name', help='Long release name (eg: "Fedora 20")'),
    click.option('--id-prefix', help='Release prefix (eg: FEDORA)'),
    click.option('--version', help='Release version number (eg: 20)'),
    click.option('--branch', help='Git branch name (eg: f20)'),
    click.option('--dist-tag', help='Koji dist tag (eg: f20)'),
    click.option('--stable-tag', help='Koji stable tag (eg: f20-updates)'),
    click.option('--released-on', type=click.DateTime(formats=["%Y-%m-%d"]),
                 help='Date of first release (eg. 2016-06-14)'),
    click.option('--eol', type=click.DateTime(formats=["%Y-%m-%d"]),
                 help='Release end-of-life date (eg. 2016-06-14)'),
    click.option('--testing-tag',
                 help='Koji testing tag (eg: f20-updates-testing)'),
    click.option('--candidate-tag',
                 help='Koji candidate tag (eg: f20-updates-candidate)'),
    click.option('--pending-stable-tag',
                 help='Koji pending tag (eg: f20-updates-pending)'),
    click.option('--pending-testing-tag',
                 help='Koji pending testing tag (eg: f20-updates-pending-testing)'),
    click.option('--pending-signing-tag',
                 help='Koji pending signing tag (eg: f20-updates-pending-signing)'),
    click.option('--override-tag', help='Koji override tag (eg: f20-override)'),
    click.option('--state', type=click.Choice(['disabled', 'pending', 'current',
                                               'frozen', 'archived']),
                 help='The state of the release'),
    click.option('--mail-template', help='Name of the email template for this release'),
    click.option('--composed-by-bodhi/--not-composed-by-bodhi', is_flag=True, default=None,
                 help='The flag that indicates whether the release is composed by Bodhi or not'),
    click.option('--package-manager', type=click.Choice(['unspecified', 'dnf', 'yum']),
                 help='The package manager used by this release'),
    click.option('--testing-repository',
                 help='The name of the testing repository used to test updates'),
    click.option(
        '--create-automatic-updates/--no-create-automatic-updates',
        help=('Configure for this release, whether or not automatic updates are '
              'created for builds which are tagged into its Koji candidate tag.'),
        is_flag=True, default=None),
    staging_option,
    url_option,
    debug_option]


def add_options(options: typing.Sequence[typing.Callable[[typing.Callable], typing.Callable]]) \
        -> typing.Callable[[typing.Callable], typing.Callable]:
    """
    Generate a click.option decorator with the given options.

    Given a list of click options this creates a decorator that
    will return a function used to add the options to a click command.

    Args:
        options: A list of click.options decorators.
    Returns:
        A decorator that applies the given options to it decorated function.
    """
    def _add_options(func: typing.Callable) -> typing.Callable:
        """
        Decorate func with the given click options.

        Given a click command and a list of click options this will
        return the click command decorated with all the options in the list.

        Args:
            func (callable): A click command function.
        Returns:
            callable: A wrapped version of func with added options.
        """
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


def handle_errors(method: typing.Callable) -> typing.Callable:
    """
    Echo neat error messages on BodhiClientException.

    This is intended to be used as a decorator on method.

    Args:
        method: The method we wish to handle errors from.
    Returns:
        A wrapped version of method that handles errors.
    """
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        """
        Call method with given args and kwargs, handling errors and exit if any are raised.

        BodhiClientExceptions will cause an exit code of 2.
        Args:
            args: Positional arguments to pass to method.
            kwargs: Keyword arguments to pass to method.
        """
        try:
            method(*args, **kwargs)
        except bindings.BodhiClientException as e:
            click.secho(str(e), fg='red', bold=True, err=True)
            sys.exit(2)
    return wrapper


def _save_override(url: str, staging: bool, edit: bool = False,
                   client_id: typing.Optional[str] = None,
                   id_provider: typing.Optional[str] = None,
                   **kwargs):
    """
    Create or edit a buildroot override.

    Args:
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        user: The username to authenticate as.
        password: The user's password.
        staging: Whether to use the staging server or not.
        edit: Set to True to edit an existing buildroot override.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=staging
    )

    resp = client.save_override(nvr=kwargs['nvr'],
                                duration=kwargs['duration'],
                                notes=kwargs['notes'],
                                edit=edit,
                                expired=kwargs.get('expire', False))

    if kwargs.get('expire', False):
        print_resp(resp, client, override_hint=False)
    elif kwargs['wait']:
        print_resp(resp, client, override_hint=False)
        command = _generate_wait_repo_command(resp, client)
        if command:
            click.echo(f"\n\nRunning {' '.join(command)}\n")
            ret = subprocess.call(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if ret:
                click.echo(f"WARNING: ensuring active override failed for {resp.build.nvr}",
                           err=True)
                sys.exit(ret)
    else:
        print_resp(resp, client, override_hint=True)


@click.group(help="Command line tool for interacting with Bodhi.")
@click.version_option(message='%(version)s', package_name='bodhi_client')
def cli():
    """Create the main CLI group."""
    pass  # pragma: no cover


@cli.group(help="Interact with composes.")
def composes():
    """Create the composes group."""
    pass  # pragma: no cover


@composes.command(name='info')
@handle_errors
@click.argument('release')
@click.argument('request')
@url_option
@add_options(openid_options)
@debug_option
@staging_option
def info_compose(release: str, request: str, url: str, id_provider: str, client_id: str, **kwargs):
    """Retrieve and print info about a compose."""
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    try:
        resp = client.get_compose(release, request)
    except bindings.ComposeNotFound as exc:
        raise click.BadParameter(str(exc), param_hint='RELEASE/REQUEST')

    print_resp(resp, client)


@composes.command(name='list')
@handle_errors
@staging_option
@click.option('-v', '--verbose', is_flag=True, default=False, help='Display more information.')
@url_option
@add_options(openid_options)
@debug_option
def list_composes(url: str, id_provider: str, client_id: str, staging: bool, verbose: bool,
                  debug: bool):
    # User docs for the CLI
    """
    List composes.

    Asterisks next to composes indicate that they contain security updates.
    """
    # developer docs
    """
    Args:
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        staging: Whether to use the staging server or not.
        verbose: Whether to show verbose output or not.
        debug: If the --debug flag was set
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=staging
    )
    print_resp(client.list_composes(), client, verbose)


@cli.group(help="Interact with updates on Bodhi.")
def updates():
    """Create the updates group."""
    pass  # pragma: no cover


def require_severity_for_security_update(type: str, severity: str):
    """
    Print an error message if the user did not provide severity for a security update.

    Args:
        type: The value of the update 'type'.
        severity: The value of the update 'severity'.
    """
    if type == 'security' and severity not in ('low', 'medium', 'high', 'urgent'):
        raise click.BadParameter('must specify severity for a security update',
                                 param_hint='severity')


@updates.command()
@click.option('--type', default='bugfix', help='Update type', required=True,
              type=click.Choice(constants.UPDATE_TYPES))
@add_options(new_edit_options)
@click.option('--from-tag', is_flag=True,
              help='Use builds from a Koji tag instead of specifying them individually')
@click.argument('builds_or_tag')
@click.option('--file', help='A text file containing all the update details')
@handle_errors
@add_options(openid_options)
@url_option
@debug_option
def new(url: str, id_provider: str, client_id: str, debug: bool, **kwargs):
    # User Docs that show in the --help
    """
    Create a new update.

    BUILDS_OR_TAG: a comma separated list of Builds to be added to the update
    (e.g. 0ad-0.0.21-4.fc26,2ping-3.2.1-4.fc26) or a single side-tag name
    """
    # Developer Docs
    """
    Args:
        user: The username to authenticate as.
        password: The user's password.
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        debug: If the --debug flag was set
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    # Because some flag options are common between new and edit, we cannot set the default
    # to be a boolean, instead we set them here
    for option in ['close_bugs', 'autotime', 'autokarma']:
        if kwargs[option] is None:
            kwargs[option] = False

    # Because bodhi.server.services.updates expects from_tag to be string
    # copy builds to from_tag and remove builds
    if kwargs['from_tag']:
        if len(kwargs['builds_or_tag'].split(' ')) > 1:
            click.echo("ERROR: Can't specify more than one tag.", err=True)
            sys.exit(1)
        kwargs['from_tag'] = kwargs.pop('builds_or_tag')
    else:
        kwargs['builds'] = kwargs.pop('builds_or_tag')
        del kwargs['from_tag']

    if kwargs['file'] is None:
        updates = [kwargs]

    else:
        updates = client.parse_file(os.path.abspath(kwargs['file']))

    kwargs['notes'] = _get_notes(**kwargs)

    if not kwargs['notes'] and not kwargs['file']:
        click.echo("ERROR: must specify at least one of --file, --notes, or --notes-file", err=True)
        sys.exit(1)

    for update in updates:
        require_severity_for_security_update(type=update['type'], severity=update['severity'])
        try:
            resp = client.save(**update)
            print_resp(resp, client)
        except bindings.BodhiClientException as e:
            click.echo(str(e), err=True)
        except Exception:
            traceback.print_exc()


def _validate_edit_update(
        ctx: click.core.Context, param: typing.Union[click.core.Option, click.core.Parameter],
        value: str) -> str:
    """
    Validate the update argument given to the updates edit command.

    The update argument can only be an update id.

    Args:
        ctx: The click Context. Unused.
        param: The name of the parameter being validated. Unused.
        value: The value of the value being validated.
    Returns:
        The value if it passes validation.
    Raises:
        click.BadParameter: If the value is invalid.
    """
    if re.search(bindings.UPDATE_ID_RE, value):
        return value
    else:
        raise click.BadParameter("Please provide an Update ID")


@updates.command()
@click.option('--type', help='Update type',
              type=click.Choice(constants.UPDATE_TYPES))
@click.option('--addbuilds', help='Add Comma-separated list of build nvr')
@click.option('--removebuilds', help='Remove Comma-separated list of build nvr')
@add_options(new_edit_options)
@click.argument('update', callback=_validate_edit_update)
@add_options(openid_options)
@url_option
@debug_option
@handle_errors
def edit(url: str, id_provider: str, client_id: str, debug: bool, **kwargs):
    # User Docs that show in the --help
    """
    Edit an existing update.

    UPDATE: The alias of the update (e.g. FEDORA-2017-f8e0ef2850)
    """
    # Developer Docs
    """
    The update argument must be an update id.

    Args:
        user: The username to authenticate as.
        password: The user's password.
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        debug: If the --debug flag was set
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    kwargs['notes'] = _get_notes(**kwargs)

    try:
        query_param = {'updateid': kwargs['update']}
        resp = client.query(**query_param)
        del kwargs['update']

        # Convert list of 'Bug' instances in DB to comma separated bug_ids for parsing.
        former_update = resp['updates'][0].copy()

        # If these flags are not set by the user we load values from the existing update
        for option in ['close_bugs', 'autotime', 'autokarma']:
            if kwargs[option] is None:
                kwargs[option] = former_update[option]

        if not kwargs['bugs']:
            kwargs['bugs'] = ",".join([str(bug['bug_id']) for bug in former_update['bugs']])
            former_update.pop('bugs', None)

        kwargs['builds'] = [b['nvr'] for b in former_update['builds']]
        kwargs['edited'] = former_update['alias']

        if former_update.get('from_tag', None):
            # The build list is always refreshed from the side-tag by validate_from_tag()
            if (kwargs['addbuilds'] or kwargs['removebuilds']):
                click.echo(
                    "ERROR: The --addbuilds and --removebuilds options"
                    " cannot be used with a side-tag update.", err=True
                )
                sys.exit(1)
            kwargs['from_tag'] = former_update['from_tag']
            del kwargs['builds']
        else:
            if kwargs['addbuilds']:
                for build in kwargs['addbuilds'].split(','):
                    if build not in kwargs['builds']:
                        kwargs['builds'].append(build)
            if kwargs['removebuilds']:
                for build in kwargs['removebuilds'].split(','):
                    kwargs['builds'].remove(build)

        del kwargs['addbuilds']
        del kwargs['removebuilds']

        # Replace empty fields with former values from database.
        for field in kwargs:
            if kwargs[field] in (None, '') and field in former_update:
                kwargs[field] = former_update[field]

        require_severity_for_security_update(type=kwargs['type'], severity=kwargs['severity'])

        resp = client.save(**kwargs)
        print_resp(resp, client)
    except bindings.BodhiClientException as e:
        click.echo(str(e), err=True)


@updates.command()
@click.option('--updateid', help='Query by update ID (eg: FEDORA-2015-0001)')
@click.option('--alias', help='Query by alias')
@click.option('--approved-since', help='Approved after a specific timestamp')
@click.option('--approved-before', help='Approved before a specific timestamp')
@click.option('--modified-since', help='Modified after a specific timestamp')
@click.option('--modified-before', help='Modified before a specific timestamp')
@click.option('--builds', help='Query updates based on builds')
@click.option('--bugs', help='A list of bug numbers')
@click.option('--critpath', is_flag=True, default=None,
              help='Query only critical path packages')
@click.option('--packages', help='Query by package name(s)')
@click.option('--content-type', help='Query updates based on content type',
              type=click.Choice(['rpm', 'module']))  # And someday, container.
@click.option('--from-side-tag/--not-from-side-tag', default=None,
              help='List only updates created from side-tag / not from side-tag')
@click.option('--gating', help='Filter by gating status', default=None,
              type=click.Choice(['passed', 'failed', 'ignored', 'waiting',
                                 'running', 'queued', 'greenwave_failed']))
@click.option('--pushed', is_flag=True, default=None,
              help='Filter by pushed updates')
@click.option('--pushed-since',
              help='Updates that have been pushed after a certain time')
@click.option('--pushed-before',
              help='Updates that have been pushed before a certain time')
@click.option('--releases', help='Updates for specific releases')
@click.option('--locked', help='Updates that are in a locked state')
@click.option('--request', help='Updates with a specific request',
              type=click.Choice(constants.REQUEST_TYPES))
@click.option('--severity', help='Updates with a specific severity',
              type=click.Choice(['unspecified', 'urgent', 'high', 'medium', 'low']))
@click.option('--submitted-since',
              help='Updates that have been submitted since a certain time')
@click.option('--submitted-before',
              help='Updates that have been submitted before a certain time')
@click.option('--status', help='Filter by update status',
              type=click.Choice(['pending', 'testing', 'stable', 'obsolete',
                                 'unpushed']))
@click.option('--suggest', help='Filter by post-update user suggestion',
              type=click.Choice(constants.SUGGEST_TYPES))
@click.option('--type', default=None, help='Filter by update type',
              type=click.Choice(constants.UPDATE_TYPES))
@click.option('--user', help='Updates submitted by a specific user')
@click.option('--mine', is_flag=True, help='Show only your updates')
@add_options(openid_options)
@staging_option
@url_option
@debug_option
@add_options(pagination_options)
@handle_errors
def query(
    url: str, id_provider: str, client_id: str, debug: bool, mine: bool = False,
    rows: typing.Optional[int] = None, **kwargs
):
    # User Docs that show in the --help
    """Query updates on Bodhi.

    A leading '*' means that this is a 'security' update.

    The number between brackets next to the date indicates the number of days
    the update is in the current state.
    """
    # Developer Docs
    """
    Query updates based on flags.

    Args:
        url: The URL of a Bodhi server to create the update on. Ignored if staging is
                       True.
        debug: If the --debug flag was set
        mine: If the --mine flag was set
        rows: How many rows to fetch
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url,
        client_id=client_id,
        id_provider=id_provider,
        staging=kwargs['staging']
    )
    if mine:
        client.ensure_auth()
        kwargs['user'] = client.username
    resp = client.query(rows_per_page=rows, **kwargs)
    print_resp(resp, client)


@updates.command()
@click.argument('update')
@click.argument('state')
@add_options(openid_options)
@staging_option
@url_option
@debug_option
@handle_errors
def request(update: str, state: str, url: str, id_provider: str, client_id: str, **kwargs):
    # User Docs that show in the --help
    """
    Change an update's request status.

    UPDATE: The alias of the update (e.g. FEDORA-2017-f8e0ef2850)

    STATE: The state you wish to change the update's request to. Valid options are
    testing, stable, obsolete, unpush, and revoke.
    """
    # Developer Docs
    """
    Change an update's request to the given state.

    Args:
        update: The update you wish to modify.
        state: The state you wish to change the update's request to. Valid options are
               testing, stable, obsolete, unpush, and revoke.
        staging: Whether to use the staging server or not.
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    try:
        resp = client.request(update, state)
    except bindings.UpdateNotFound as exc:
        raise click.BadParameter(str(exc), param_hint='UPDATE')

    print_resp(resp, client)


@updates.command()
@click.argument('update')
@click.argument('text')
@click.option('--karma', default=0, type=click.INT, help='The karma for this comment (+1/0/-1)')
@add_options(openid_options)
@staging_option
@url_option
@debug_option
@handle_errors
def comment(update: str, text: str, karma: int, url: str, id_provider: str, client_id: str,
            **kwargs):
    # User Docs that show in the --help
    """
    Comment on an update.

    UPDATE: The alias of the update (e.g. FEDORA-2017-f8e0ef2850)

    TEXT: the comment to be added to the update
    """
    # Developer Docs
    """
    Comment on an update.

    Args:
        update: The update you wish to modify.
        text: The text of the comment you wish to leave on the update.
        karma: The karma you wish to leave on the update. Must be +1, 0, or -1.
        staging: Whether to use the staging server or not.
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """

    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )
    resp = client.comment(update, text, karma)
    print_resp(resp, client)


@updates.command()
@staging_option
@click.option('--arch',
              help=('Specify arch of packages to download, "all" will retrieve packages from all '
                    'architectures'))
@click.option('--debuginfo', is_flag=True, default=False,
              help=('Include debuginfo packages'))
@click.option('--updateid', help='Download update(s) by ID(s) (comma-separated list)')
@click.option('--builds', help='Download update(s) by build NVR(s) (comma-separated list)')
@url_option
@add_options(openid_options)
@debug_option
@handle_errors
def download(url: str, id_provider: str, client_id: str, **kwargs):
    # User Docs that show in the --help
    """Download the builds in one or more updates."""
    # Developer Docs
    """
    Download the builds for an update.

    Args:
        staging:   Whether to use the staging server or not.
        arch:   Requested architecture of packages to download.
                "all" will retrieve packages from all architectures.
        debuginfo: Whether to include debuginfo packages.
        url:    The URL of a Bodhi server to create the update on. Ignored if staging is True.
        kwargs:    Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )
    requested_arch = kwargs['arch']
    debuginfo = kwargs['debuginfo']

    del kwargs['staging']
    del kwargs['arch']
    del kwargs['debuginfo']
    # At this point we need to have reduced the kwargs dict to only our
    # query options (updateid or builds)
    if not any(kwargs.values()):
        click.echo("ERROR: must specify at least one of --updateid or --builds", err=True)
        sys.exit(1)

    # As the query method doesn't let us construct OR queries, we're
    # gonna run one query for each option that was passed. The syntax
    # for this is a bit ugly, sorry.
    for (attr, value) in kwargs.items():
        if value:
            expecteds = len(value.split(','))
            resp = client.query(**{attr: value})
            if len(resp.updates) == 0:
                click.echo(f"WARNING: No {attr} found!", err=True)
            else:
                if attr == 'updateid':
                    resp_no = len(resp.updates)
                else:
                    # attr == 'builds', add up number of builds for each returned update
                    resp_no = functools.reduce(
                        lambda x, y: x + len(y.get('builds', [])), resp.updates, 0
                    )

                if resp_no < expecteds:
                    click.echo(f"WARNING: Some {attr} not found!", err=True)
                # Not sure if we need a check for > expecteds, I don't
                # *think* that should ever be possible for these opts.

            for update in resp.updates:
                click.echo(f"Downloading packages from {update['alias']}")
                for build in update['builds']:
                    args = ['koji', 'download-build']
                    if debuginfo:
                        args.append('--debuginfo')
                    # subprocess is icky, but koji module doesn't
                    # expose this in any usable way, and we don't want
                    # to rewrite it here.
                    if requested_arch is None:
                        args.extend(['--arch=noarch',
                                     f'--arch={platform.machine()}', build['nvr']])
                    else:
                        if 'all' in requested_arch:
                            args.append(build['nvr'])
                        if 'all' not in requested_arch:
                            args.extend(['--arch=noarch',
                                         f'--arch={requested_arch}', build['nvr']])
                    ret = subprocess.call(args)
                    if ret:
                        click.echo(f"WARNING: download of {build['nvr']} failed!", err=True)


def _get_notes(**kwargs) -> str:
    """
    Return notes for the update.

    If the user provides a --notes-file, _get_notes processes the contents of the notes-file.
    If the user does not provide a --notes-file, _get_notes() returns the notes from the kwargs.
    One cannot specify both --notes and --notesfile. Doing so will result in an error.

    Args:
        kwargs: Keyword arguments passed to us by click.
    Returns:
        The contents of the notes file or the notes from kwargs
    """
    if kwargs['notes_file'] is not None:
        if kwargs['notes'] is None:
            with open(kwargs['notes_file'], 'r') as fin:
                return fin.read()
        else:
            click.echo("ERROR: Cannot specify --notes and --notes-file", err=True)
            sys.exit(1)
    else:
        return kwargs['notes']


@updates.command()
@click.argument('update')
@click.argument('comment', required=False)
@click.option(
    '--show', is_flag=True, default=None,
    help="List all the required unsatisfied requirements")
@click.option(
    '--test', multiple=True,
    help="Waive the specific test(s), to automatically waive all unsatisfied "
    "requirements, specify --test=all")
@add_options(openid_options)
@staging_option
@url_option
@debug_option
@handle_errors
def waive(update: str, show: bool, test: typing.Iterable[str], comment: str, url: str,
          id_provider: str, client_id: str, **kwargs):
    # User Docs that show in the --help
    """
    Show or waive unsatified requirements (ie: missing or failing tests) on an existing update.

    UPDATE: The alias of the update (e.g. FEDORA-2017-f8e0ef2850)

    COMMENT: A comment explaining why the requirements were waived (mandatory with --test)
    """
    # Developer Docs
    """
    The update argument must be an update id..

    Args:
        update: The update who unsatisfied requirements wish to waive.
        show: Whether to show all missing required tests of the specified update.
        test: Waive those specified tests or all of them if 'all' is specified.
        comment: A comment explaining the waiver.
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    if show and test:
        click.echo(
            'ERROR: You can not list the unsatisfied requirements and waive them '
            'at the same time, please use either --show or --test=... but not both.',
            err=True)
        sys.exit(1)

    if show:
        test_status = client.get_test_status(update)
        if 'errors' in test_status:
            click.echo('One or more errors occurred while retrieving the unsatisfied requirements:',
                       err=True)
            for el in test_status.errors:
                click.echo(f'  - {el.description}', err=True)
        elif 'decision' not in test_status:
            click.echo('Could not retrieve the unsatisfied requirements from bodhi.', err=True)
        else:
            click.echo(f'CI status: {test_status.decision.summary}')
            if test_status.decision.unsatisfied_requirements:
                click.echo('Missing tests:')
                for req in test_status.decision.unsatisfied_requirements:
                    click.echo(f'  - {req.testcase}')
            else:
                click.echo('Missing tests: None')
    else:
        if not comment:
            click.echo('ERROR: A comment is mandatory when waiving unsatisfied requirements',
                       err=True)
            sys.exit(1)

        if 'all' in test:
            click.echo('Waiving all unsatisfied requirements')
            resp = client.waive(update, comment)
        else:
            click.echo(f"Waiving unsatisfied requirements: {', '.join(test)}")
            resp = client.waive(update, comment, test)
        print_resp(resp, client)


@updates.command(name="trigger-tests")
@click.argument('update')
@add_options(openid_options)
@staging_option
@url_option
@debug_option
@handle_errors
def trigger_tests(update: str, url: str, id_provider: str, client_id: str, **kwargs):
    # User Docs that show in the --help
    """
    Trigger tests for existing update in testing status.

    UPDATE: The alias of the update (e.g. FEDORA-2017-f8e0ef2850)
    """
    # Developer Docs
    """
    The update argument must be an update id..

    Args:
        update: The update we want to trigger tests for.
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    resp = client.trigger_tests(update)
    print_resp(resp, client)


@cli.group(help="Interact with overrides on Bodhi.")
def overrides():
    """Create the overrides CLI group."""
    pass  # pragma: no cover


@overrides.command('query')
@click.option('--user', default=None,
              help='Overrides submitted by a specific user')
@staging_option
@click.option('--mine', is_flag=True,
              help='Show only your overrides.')
@click.option('--packages', default=None,
              help='Query by comma-separated package name(s)')
@click.option('--expired/--active', default=None,
              help='show only expired or active overrides')
@click.option('--releases', default=None,
              help='Query by release shortname(s). e.g. F26')
@click.option('--builds', default=None,
              help='Query by comma-separated build id(s)')
@url_option
@add_options(openid_options)
@debug_option
@add_options(pagination_options)
@handle_errors
def query_buildroot_overrides(
        url: str, id_provider: str, client_id: str,
        user: typing.Optional[str] = None, mine: bool = False,
        packages: typing.Optional[str] = None, expired: typing.Optional[bool] = None,
        releases: typing.Optional[str] = None, builds: typing.Optional[str] = None,
        rows: typing.Optional[int] = None, page: typing.Optional[int] = None, **kwargs):
    # Docs that show in the --help
    """Query the buildroot overrides."""
    # Developer Docs
    """
    Query the buildroot overrides.

    Args:
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        user: If supplied, overrides for this user will be queried.
        mine: Whether to use the --mine flag was given.
        packages: If supplied, the overrides for these package are queried
        expired: If supplied, True returns only expired overrides, False only active.
        releases: If supplied, the overrides for these releases are queried.
        builds: If supplied, the overrides for these builds are queried.
        rows: The limit of rows displayed per page for query result.
        page: If supplied, returns the results for a specific page number.
        kwargs: Other keyword arguments passed to us by click.
    """
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )
    if mine:
        client.ensure_auth()
        user = client.username
    resp = client.list_overrides(user=user, packages=packages,
                                 expired=expired, releases=releases, builds=builds,
                                 rows_per_page=rows, page=page)
    print_resp(resp, client)


@overrides.command('save')
@add_options(save_edit_options)
@add_options(openid_options)
@handle_errors
def save_buildroot_overrides(url: str, id_provider: str, client_id: str, staging: bool,
                             **kwargs):
    # Docs that show in the --help
    """
    Create a buildroot override.

    NVR: the NVR (name-version-release) of the buildroot override to create
    """
    # Developer Docs
    """
    Create a buildroot override.

    Args:
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        staging: Whether to use the staging server or not.
        kwargs: Other keyword arguments passed to us by click.
    """
    _save_override(url=url, client_id=client_id, id_provider=id_provider, staging=staging,
                   **kwargs)


@overrides.command('edit')
@add_options(save_edit_options)
@add_options(openid_options)
@click.option('--expire', help='Expire the override', is_flag=True, default=False)
@handle_errors
def edit_buildroot_overrides(url: str, id_provider: str, client_id: str, staging: bool, **kwargs):
    # Docs that show in the --help
    """
    Edit a buildroot override.

    NVR: the NVR (name-version-release) of the buildroot override to edit
    """
    # Developer Docs
    """
    Edit a buildroot override.

    Args:
        url: The URL of a Bodhi server to create the update on. Ignored if staging is True.
        client_id: The OIDC client_id.
        id_provider: The OIDC provider.
        staging: Whether to use the staging server or not.
        kwargs: Other keyword arguments passed to us by click.
    """
    _save_override(url=url, client_id=client_id, id_provider=id_provider, staging=staging,
                   edit=True, **kwargs)


def _generate_wait_repo_command(override: munch.Munch, client: bindings.BodhiClient) \
        -> typing.Optional[typing.Tuple[str, str, str, str]]:
    """
    Generate and return a koji wait-repo command for the given override, if possible.

    Args:
        override: A Munch of the Override we want to print a hint about.
        client: A BodhiClient that we can use to query the server for Releases.
    Returns:
        If we know the release for the override's build, we return a tuple suitable for passing to
        subprocess.Popen for a koji command that will wait on the repo. If we can't we return None.
    """
    if 'release_id' in override.build:
        release = client.get_releases(ids=[override.build.release_id])['releases'][0]
        return ('koji', 'wait-repo', f'{release.dist_tag}-build',
                f'--build={override.build.nvr}')
    return None


def _print_override_koji_hint(override: munch.Munch, client: bindings.BodhiClient):
    """
    Print a human readable hint about how to use koji wait-repo to monitor an override, if possible.

    Note: The hint can only be generated if the server provides a 'release_id' on the
    override.build property. Older versions of the server did not include the release_id on
    Build objects during serialization, and those server versions also did not allow querying
    for releases by id. If override.build.release_id is not found, None will be returned.

    Args:
        override: A Munch of the Override we want to print a hint about.
        client: A BodhiClient that we can use to query the server for Releases.
    """
    command = _generate_wait_repo_command(override, client)
    if command:
        click.echo(
            '\n\nUse the following to ensure the override is active:\n\n'
            f"\t$ {' '.join(command)}\n")


def print_resp(resp: munch.Munch, client: bindings.BodhiClient, verbose: bool = False,
               override_hint: bool = True):
    """
    Print a human readable rendering of the given server response to the terminal.

    Args:
        resp: The response from the server.
        client: A BodhiClient.
        verbose: If True, show more detailed output. Defaults to False.
        override_hint: If True, show a hint to the user about how to wait on a buildroot
            override. Defaults to True.
    """
    if 'updates' in resp:
        if len(resp.updates) == 1:
            click.echo(client.update_str(resp.updates[0]))
        else:
            for update in resp.updates:
                click.echo(client.update_str(update, minimal=True))
        if 'total' in resp:
            click.echo(f'{resp.total} updates found ({len(resp.updates)} shown)')
    elif resp.get('update'):
        click.echo(client.update_str(resp['update']))
    elif resp.get('alias'):
        click.echo(client.update_str(resp))
    elif 'overrides' in resp:
        if len(resp.overrides) == 1:
            click.echo(client.override_str(resp.overrides[0], minimal=False))
            if override_hint:
                _print_override_koji_hint(resp.overrides[0], client)
        else:
            for override in resp.overrides:
                click.echo(client.override_str(override).strip())
        if 'total' in resp:
            click.echo(
                f'{resp.total} overrides found ({len(resp.overrides)} shown)')
    elif 'build' in resp:
        click.echo(client.override_str(resp, minimal=False))
        if override_hint:
            _print_override_koji_hint(resp, client)
    elif 'comment' in resp:
        click.echo(f"The following comment was added to {resp.comment['update'].alias}")
        click.echo(resp.comment.text)
    elif 'compose' in resp:
        click.echo(client.compose_str(resp['compose'], minimal=False))
    elif 'composes' in resp:
        if len(resp['composes']) == 1:
            click.echo(client.compose_str(resp['composes'][0], minimal=(not verbose)))
        else:
            for compose in resp['composes']:
                click.echo(client.compose_str(compose, minimal=(not verbose)))
                if verbose:
                    # Let's add a little more spacing
                    click.echo()
    else:
        click.echo(resp)
    if resp.get('caveats', None):
        click.echo('Caveats:')
        for caveat in resp.caveats:
            click.echo(caveat.description)


@cli.group(help="Interact with releases.")
def releases():
    """Manage the releases."""
    pass  # pragma: no cover


@releases.command(name='create')
@handle_errors
@add_options(release_options)
@add_options(openid_options)
def create_release(url: str, id_provider: str, client_id: str, debug: bool, **kwargs):
    """Create a release."""
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )
    kwargs['csrf_token'] = client.csrf()
    # the defaults for these are set to None so that edit_release
    # does not change the current value unless it was passed on the
    # command line; for creating a new release, we set the defaults
    # here
    if kwargs['composed_by_bodhi'] is None:
        kwargs['composed_by_bodhi'] = True
    if kwargs['create_automatic_updates'] is None:
        kwargs['create_automatic_updates'] = False
    kwargs['released_on'] = kwargs.pop('released_on').date() if kwargs['released_on'] else None
    kwargs['eol'] = kwargs.pop('eol').date() if kwargs['eol'] else None
    save(client, **kwargs)


@releases.command(name='edit')
@handle_errors
@add_options(release_options)
@add_options(openid_options)
@click.option('--new-name', help='New release name (eg: F20)')
def edit_release(url: str, id_provider: str, client_id: str, debug: bool, **kwargs):
    """Edit an existing release."""
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )
    csrf = client.csrf()

    edited = kwargs.pop('name')

    kwargs['released_on'] = kwargs.pop('released_on').date() if kwargs['released_on'] else None
    kwargs['eol'] = kwargs.pop('eol').date() if kwargs['eol'] else None

    if edited is None:
        click.echo("ERROR: Please specify the name of the release to edit", err=True)
        return

    res = client.send_request(f'releases/{edited}', verb='GET', auth=True)

    data = munch.unmunchify(res)

    if 'errors' in data:
        print_errors(data)

    data['edited'] = edited
    data['csrf_token'] = csrf

    new_name = kwargs.pop('new_name')

    if new_name is not None:
        data['name'] = new_name

    if (data['state'] == 'pending' and kwargs['state'] == 'current'
            and kwargs['released_on'] is None):
        kwargs['released_on'] = datetime.now().date()

    for k, v in kwargs.items():
        if v is not None:
            data[k] = v

    save(client, **data)


@releases.command(name='info')
@handle_errors
@click.argument('name')
@url_option
@add_options(openid_options)
@debug_option
@staging_option
def info_release(name: str, url: str, id_provider: str, client_id: str, **kwargs):
    """Retrieve and print info about a named release."""
    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    res = client.send_request(f'releases/{name}', verb='GET', auth=False)

    if 'errors' in res:
        print_errors(res)

    else:
        click.echo('Release:')
        print_release(res)


@releases.command(name='list')
@handle_errors
@click.option('--display-archived', is_flag=True, default=False,
              help='Display full list, including archived releases.')
@url_option
@add_options(openid_options)
@debug_option
@add_options(pagination_options)
@staging_option
def list_releases(display_archived: bool, url: str, id_provider: str, client_id: str,
                  rows: typing.Optional[int] = None, page: typing.Optional[int] = None, **kwargs):
    """Retrieve and print list of releases."""
    exclude_archived = True
    if display_archived:
        exclude_archived = False

    client = bindings.BodhiClient(
        base_url=url, client_id=client_id, id_provider=id_provider, staging=kwargs['staging']
    )

    res = client.get_releases(rows_per_page=rows, page=page, exclude_archived=exclude_archived)

    print_releases_list(res['releases'])


def save(client: bindings.BodhiClient, **kwargs):
    """
    Save a new or edited release.

    Args:
        client: The Bodhi client to use for the request.
        kwargs: The parameters to send with the request.
    """
    res = client.send_request('releases/', verb='POST', auth=True,
                              data=kwargs)

    if 'errors' in res:
        print_errors(res)

    else:
        click.echo("Saved release:")
        print_release(res)


def print_releases_list(releases: typing.List[munch.Munch]):
    """
    Print a list of releases to the terminal.

    Args:
        releases (munch.Munch): The releases to be printed.
    """
    pending = [release for release in releases if release['state'] == 'pending']
    archived = [release for release in releases if release['state'] == 'archived']
    current = [release for release in releases if release['state'] == 'current']
    frozen = [release for release in releases if release['state'] == 'frozen']

    if pending:
        click.echo('pending:')
        for release in pending:
            click.echo(f"  Name:                {release['name']}")

    if archived:
        click.echo('\narchived:')
        for release in archived:
            click.echo(f"  Name:                {release['name']}")

    if current:
        click.echo('\ncurrent:')
        for release in current:
            click.echo(f"  Name:                {release['name']}")

    if frozen:
        click.echo('\nfrozen:')
        for release in frozen:
            click.echo(f"  Name:                {release['name']}")


def print_release(release: munch.Munch):
    """
    Print a given release to the terminal.

    Args:
        release (munch.Munch): The release to be printed.
    """
    click.echo(f"  Name:                     {release['name']}")
    click.echo(f"  Long Name:                {release['long_name']}")
    click.echo(f"  Version:                  {release['version']}")
    click.echo(f"  Branch:                   {release['branch']}")
    click.echo(f"  ID Prefix:                {release['id_prefix']}")
    click.echo(f"  Dist Tag:                 {release['dist_tag']}")
    click.echo(f"  Stable Tag:               {release['stable_tag']}")
    click.echo(f"  Testing Tag:              {release['testing_tag']}")
    click.echo(f"  Candidate Tag:            {release['candidate_tag']}")
    click.echo(f"  Pending Signing Tag:      {release['pending_signing_tag']}")
    click.echo(f"  Pending Testing Tag:      {release['pending_testing_tag']}")
    click.echo(f"  Pending Stable Tag:       {release['pending_stable_tag']}")
    click.echo(f"  Override Tag:             {release['override_tag']}")
    click.echo(f"  State:                    {release['state']}")
    click.echo(f"  Beta Status:              {release['setting_status']}")
    click.echo(f"  Email Template:           {release['mail_template']}")
    click.echo(f"  Composed by Bodhi:        {release['composed_by_bodhi']}")
    click.echo(f"  Create Automatic Updates: {release['create_automatic_updates']}")
    click.echo(f"  Package Manager:          {release['package_manager']}")
    click.echo(f"  Testing Repository:       {release['testing_repository']}")
    click.echo(f"  Released On:              {release['released_on']}")
    click.echo(f"  End of Life:              {release['eol']}")


def print_errors(data: munch.Munch):
    """
    Print errors to the terminal and exit with code 1.

    Args:
        data (munch.Munch): The errors to be formatted and printed.
    """
    for error in data['errors']:
        click.echo(f"ERROR: {error['description']}", err=True)

    sys.exit(1)


if __name__ == "__main__":
    cli()  # pragma: no cover



import threading
import time

from authlib.integrations.base_client.errors import OAuthError
from click import ClickException
from click.exceptions import Abort
import pytest
import requests

from bodhi.client.oidcclient import JSONStorage, OIDCClient, OIDCClientError

from .utils import build_response


@pytest.fixture
def storage(mocker):
    obj = mocker.Mock(name="storage")
    obj.load.return_value = {}
    return obj


@pytest.fixture
def client(mocker, storage):
    requests_mock = mocker.patch("bodhi.client.oidcclient.requests")
    response = mocker.Mock()
    response.json.return_value = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": ["query", "none"]
    }
    requests_mock.get.return_value = response
    client = OIDCClient("client_id", "scope", "http://id.example.com", storage)
    mocker.stopall()
    return client


def test_oidcclient(mocker, storage):
    metadata = {"token_endpoint": "http://id.example.com/token"}
    requests_mock = mocker.patch("bodhi.client.oidcclient.requests")
    response = mocker.Mock()
    response.json.return_value = metadata
    requests_mock.get.return_value = response
    client = OIDCClient("client_id", "scope", "http://id.example.com", storage)
    assert client.client_id == "client_id"
    assert client.scope == "scope"
    assert client.storage == storage
    requests_mock.get.assert_called_with("http://id.example.com/.well-known/openid-configuration")
    assert client.metadata == metadata


def test_oidcclient_use_oob(mocker, storage):
    requests_mock = mocker.patch("bodhi.client.oidcclient.requests")
    response = mocker.Mock()
    requests_mock.get.return_value = response

    # Test unsupported
    response.json.return_value = {
        "token_endpoint": "http://id.example.com/token",
        "response_modes_supported": ["query", "none"]
    }
    client = OIDCClient("client_id", "scope", "http://id.example.com", storage)
    assert client._use_oob is False
    assert client.redirect_uri == "http://localhost:45678/auth"
    # Test supported
    response.json.return_value = {
        "token_endpoint": "http://id.example.com/token",
        "response_modes_supported": ["query", "none", "oob"]
    }
    client = OIDCClient("client_id", "scope", "http://id.example.com", storage)
    assert client._use_oob is True
    assert client.redirect_uri == "urn:ietf:wg:oauth:2.0:oob"


def test_oidcclient_metadata_error(mocker, storage):
    requests_mock = mocker.patch("bodhi.client.oidcclient.requests")
    requests_mock.get.return_value = build_response(404, "/metadata", "not found")
    with pytest.raises(OIDCClientError) as exc:
        OIDCClient("client_id", "scope", "http://id.example.com", storage)
    requests_mock.get.assert_called_with("http://id.example.com/.well-known/openid-configuration")
    assert str(exc.value) == "not found"


def test_oidcclient_login(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    mocker.patch("bodhi.client.oidcclient.click.prompt", return_value="result-code")
    # Enable OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": ["oob"]
    }
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    oauth2client.fetch_token.return_value = "result-token"
    client.login()
    oauth2client.create_authorization_url.assert_called_with("http://id.example.com/auth")
    oauth2client.fetch_token.assert_called_with(
        'http://id.example.com/token',
        authorization_response="?result-code",
        redirect_uri='urn:ietf:wg:oauth:2.0:oob',
    )
    assert client.tokens == "result-token"


def test_oidcclient_login_with_kerberos(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    response = mocker.Mock()
    sample_code = "code=d37deb2e-5463-1234_5EH4MhV3L&amp;state=k44Rw1"
    response.text = (
        '<meta charset="UTF-8">\n<title>' + sample_code + '</title>\n   '
    )
    mocker.patch("bodhi.client.oidcclient.requests.get", return_value=response)
    # Enable OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": ["oob"]
    }
    oauth2client.fetch_token.return_value = "result-token"
    client.login_with_kerberos("auth-url")
    oauth2client.fetch_token.assert_called_with(
        'http://id.example.com/token',
        authorization_response=f"?{sample_code}",
        redirect_uri='urn:ietf:wg:oauth:2.0:oob',
    )
    assert client.tokens == "result-token"


def test_oidcclient_login_with_kerberos_wrong_code(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    response = mocker.Mock()
    sample_code = "code=NOT_4_VALID=STRING@#*"
    response.text = (
        '<meta charset="UTF-8">\n<title>' + sample_code + '</title>\n   '
    )
    mocker.patch("bodhi.client.oidcclient.requests.get", return_value=response)
    # Enable OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": ["oob"]
    }
    oauth2client.fetch_token.return_value = "result-token"
    with pytest.raises(OIDCClientError) as exc:
        client.login_with_kerberos("auth-url")
    assert str(exc.value) == 'Unable to locate OIDC code in the response from "auth-url".'


def test_oidcclient_login_with_kerberos_error_during_auth(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    response = mocker.Mock()
    response.text = "<response>"
    mocker.patch("bodhi.client.oidcclient.requests.get", return_value=response)
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": []
    }

    def raise_http_error():
        raise requests.HTTPError("error")
    response.raise_for_status = raise_http_error
    with pytest.raises(OIDCClientError) as exc:
        client.login_with_kerberos("auth-url")
    assert str(exc.value) == 'There was an issue while performing Kerberos authentication: error'


def test_oidcclient_login_kerberos_fallback(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    login_with_kerberos = mocker.patch.object(
        client, "login_with_kerberos", side_effect=OIDCClientError
    )
    login_with_browser = mocker.patch.object(client, "login_with_browser")
    client.login(use_kerberos=True)
    login_with_kerberos.assert_called_once_with("auth-url")
    login_with_browser.assert_called_once_with("auth-url")


def test_oidcclient_login_no_oob(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    # No OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": []
    }
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    oauth2client.fetch_token.return_value = "result-token"
    login_thread = threading.Thread(target=client.login)
    login_thread.start()
    # Give the embedded HTTP server some time to start up.
    time.sleep(1)
    response = requests.get(client.redirect_uri, params={"code": "CODE"})
    login_thread.join()
    assert response.ok is True
    assert "Authentication successful!" in response.text
    assert "You can now close this browser window" in response.text

    oauth2client.fetch_token.assert_called_once_with(
        'http://id.example.com/token',
        authorization_response="/auth?code=CODE",
        redirect_uri=client.redirect_uri,
    )
    assert client.tokens == "result-token"


def test_oidcclient_login_no_oob_interrupted(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    # No OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": []
    }
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    mocker.patch.object(threading.Thread, "start")
    mocker.patch.object(threading.Thread, "join", side_effect=[KeyboardInterrupt, None])
    with pytest.raises(ClickException) as exc:
        client.login()
    assert str(exc.value) == "Cancelled."


def test_oidcclient_login_no_oob_failure(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    # No OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": []
    }
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    oauth2client.fetch_token.side_effect = ValueError("not good")
    login_thread = threading.Thread(target=client.login)
    login_thread.start()
    # Give the embedded HTTP server some time to start up.
    time.sleep(1)
    response = requests.get(client.redirect_uri, params={"code": "CODE"})
    login_thread.join()
    assert response.ok is False
    assert "not good" in response.text


def test_oidcclient_login_interrupted(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    mocker.patch("bodhi.client.oidcclient.click.prompt", side_effect=Abort)
    # Enable OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": ["oob"]
    }
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    with pytest.raises(SystemExit) as exc:
        client.login()
    oauth2client.fetch_token.assert_not_called()
    assert str(exc.value) == "Cancelled."


def test_oidcclient_login_retry(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    client._tokens = None
    client.storage.load.side_effect = lambda k, d: d
    prompt = mocker.patch("bodhi.client.oidcclient.click.prompt")
    prompt.return_value = "result-code"
    secho = mocker.patch("bodhi.client.oidcclient.click.secho")
    # Enable OOB
    client.metadata = {
        "token_endpoint": "http://id.example.com/token",
        "authorization_endpoint": "http://id.example.com/auth",
        "response_modes_supported": ["oob"]
    }
    oauth2client.create_authorization_url.return_value = ("auth-url", "state")
    oauth2client.fetch_token.side_effect = [
        OAuthError("nope", "wrong"),
        OAuthError("nope", "wrong again"),
        "result-token",
    ]

    client.login()

    assert prompt.call_count == 3
    assert oauth2client.fetch_token.call_count == 3
    assert client.tokens == "result-token"
    assert secho.call_args_list == [
        mocker.call("Authenticating... Please open your browser to:", fg="yellow"),
        mocker.call("Login failed!: nope: wrong. Please try again.", fg="red"),
        mocker.call("Login failed!: nope: wrong again. Please try again.", fg="red"),
        mocker.call("Login successful!", fg="green"),
    ]


def test_oidcclient_reuse_token(mocker, client, storage):
    # Reset the tokens cache
    client._tokens = None
    client.storage.load.return_value = {"http://id.example.com": {"foo": "bar"}}
    assert client.tokens == {"foo": "bar"}
    login = mocker.patch.object(client, "login")
    client.ensure_auth()
    login.assert_not_called


def test_oidcclient_username(mocker, client):
    oauth2client = mocker.Mock()
    client.client = oauth2client
    client._tokens = {"access_token": "TOKEN"}
    client.metadata = {
        "userinfo_endpoint": "http://id.example.com/user",
    }
    response = mocker.Mock()
    response.json.return_value = {"nickname": "NICKNAME"}
    oauth2client.get.return_value = response
    assert client.username == "NICKNAME"
    oauth2client.get.assert_called_with("http://id.example.com/user")


def test_oidcclient_request(mocker, client):
    client._tokens = {"access_token": "TOKEN"}
    client.client.token_auth.set_token({"access_token": "TOKEN"})
    request = mocker.patch.object(client.client, "request")
    client.request("GET", "http://id.example.com")
    request.assert_called_with("GET", "http://id.example.com")


def test_oidcclient_request_connection_error(mocker, client):
    client._tokens = {"access_token": "TOKEN"}
    client.client.token_auth.set_token({"access_token": "TOKEN"})
    request = mocker.patch.object(client.client, "request")
    request.side_effect = requests.exceptions.ConnectionError("failed!")
    with pytest.raises(OIDCClientError) as exc:
        client.request("GET", "http://id.example.com")
    assert str(exc.value) == "failed!"


def test_oidcclient_request_auth_error(mocker, client):
    client._tokens = {"access_token": "TOKEN"}
    client.client.token_auth.set_token({"access_token": "TOKEN"})
    request = mocker.patch.object(client.client, "request")
    request.side_effect = [
        OAuthError("failed!"),
        "OK this time",
    ]
    login = mocker.patch.object(client, "login")

    result = client.request("GET", "http://id.example.com")

    assert result == "OK this time"
    assert request.call_count == 2
    login.assert_called_with(use_kerberos=False)
    assert client.tokens == {}


def test_oidcclient_clear_auth(mocker, client):
    client._tokens = {"access_token": "TOKEN"}
    client._username == "NICKNAME"
    client.client.token_auth.set_token({"access_token": "TOKEN"})
    client.client.cookies.set("foo", "bar")
    login = mocker.patch.object(client, "login")
    client.metadata = {
        "userinfo_endpoint": "http://id.example.com/user",
    }
    request = mocker.patch.object(client.client, "request")

    client.clear_auth()

    assert client.tokens == {}
    assert client.client.token_auth.token == {}
    print(client.client.cookies.list_domains())
    assert not client.has_cookie("foo")
    # Now make sure we login again if asked to.
    client.ensure_auth()
    login.assert_called
    assert client.username != "NICKNAME"
    request.assert_called_with("GET", "http://id.example.com/user", allow_redirects=True)


def test_oidcclient_has_cookie(client):
    client.client.cookies.clear()
    assert client.has_cookie("foo") is False
    client.client.cookies.set("foo", "bar", domain="localhost.local")
    assert client.has_cookie("foo") is True
    # has_cookie() should add ".local" when not fully qualified
    assert client.has_cookie("foo", domain="localhost") is True
    assert client.has_cookie("foo", domain="example.com") is False


def test_oidcclient_update_token(client):
    client._update_token({"foo": "bar"})
    assert client.tokens == {"foo": "bar"}


def test_storage(tmpdir):
    storage = JSONStorage(tmpdir.join("storage.json"))
    storage.save("tokens", {"foo": "bar"})
    assert storage.load("tokens") == {"foo": "bar"}


def test_storage_make_directory(tmpdir):
    storage = JSONStorage(tmpdir.join("subdir", "storage.json"))
    storage.save("tokens", {"foo": "bar"})
    assert storage.load("tokens") == {"foo": "bar"}
